import contextlib
import contextvars
import dataclasses
import inspect
import textwrap
import typing

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import typing_extensions

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    weave_internal,
    errors,
    _dict_utils,
    artifact_base,
    box,
    context_state,
    debug_types,
    graph,
    op_def,
    op_def_type,
    node_ref,
    ref_base,
)
from weave.legacy.weave.arrow.arrow import (
    ArrowWeaveListType,
    arrow_as_array,
    arrow_zip,
    offsets_starting_at_zero,
    pretty_print_arrow_type,
    safe_is_null,
)
from weave.legacy.weave.language_features.tagging import (
    tag_store,
    tagged_value_type,
)
from weave.trace import ref_util


def reverse_dict(d: dict) -> dict:
    return dict(reversed(d.items()))


def concrete_to_tagstore(val: typing.Any) -> typing.Any:
    if isinstance(val, dict):
        if "_tag" in val and "_value" in val:
            v = box.box(concrete_to_tagstore(val["_value"]))
            tag_store.add_tags(v, concrete_to_tagstore(val["_tag"]))
            return v
        return {k: concrete_to_tagstore(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [concrete_to_tagstore(v) for v in val]
    elif dataclasses.is_dataclass(val):
        params = {
            f.name: concrete_to_tagstore(getattr(val, f.name))
            for f in dataclasses.fields(val)
        }
        return val.__class__(**params)
    return val


def diff_arrow_type(t1, t2):
    if t1 == t2:
        return None
    if (
        isinstance(t1, pa.StructType)
        and isinstance(t2, pa.StructType)
        or isinstance(t1, pa.UnionType)
        and isinstance(t2, pa.UnionType)
    ):
        if isinstance(t1, pa.StructType):
            error_prefix = "structs"
        else:
            error_prefix = "unions"
        if t1.num_fields != t2.num_fields:
            return f"{error_prefix} different number of fields"
        for f1, f2 in zip(t1, t2):
            if f1.name != f2.name:
                return f"{error_prefix} different field names: {f1.name} != {f2.name}"
            diff = diff_arrow_type(f1.type, f2.type)
            if diff is not None:
                return f"{error_prefix} field {f1.name}: {diff}"
        return None
    if isinstance(t1, pa.ListType) and isinstance(t2, pa.ListType):
        return diff_arrow_type(t1.value_type, t2.value_type)
    return f"{t1}\n!=\n{t2}"


def first_non_none(v):
    for i in v:
        if i is not None:
            return i
    return None


@dataclasses.dataclass(frozen=True)
class PathItemOuterList:
    # Denotes this is the outermost list
    pass


@dataclasses.dataclass(frozen=True)
class PathItemStructField:
    key: str


@dataclasses.dataclass(frozen=True)
class PathItemObjectField:
    attr: str


@dataclasses.dataclass(frozen=True)
class PathItemList:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemTaggedValueTag:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemTaggedValueValue:
    pass


@dataclasses.dataclass(frozen=True)
class PathItemUnionEntry:
    type_code: int


PathItemType = typing.Union[
    PathItemOuterList,
    PathItemStructField,
    PathItemObjectField,
    PathItemList,
    PathItemTaggedValueTag,
    PathItemTaggedValueValue,
    PathItemUnionEntry,
]

PathType = tuple[PathItemType, ...]


def set_path(
    v: typing.Any,
    arrow_data: pa.Array,
    path: PathType,
    value_fn: typing.Callable[[typing.Any, int], typing.Any],
    offset: int = 0,
    arrow_offset: int = 0,
):
    last_settable_index = len(path) - 1
    final_union_path_el: typing.Optional[PathItemUnionEntry] = None
    if isinstance(path[last_settable_index], PathItemUnionEntry):
        last_settable_index -= 1
        final_union_path_el = path[-1]  # type: ignore
    if isinstance(path[last_settable_index], PathItemUnionEntry):
        raise errors.WeaveInternalError("path ended with two unions, invalid")
    # Helper function for converting arrow to python. Set the value at a path
    # within the value v, using value_fn to get the new value.
    # Path must be of length 1 or greater. Caller needs to handle the case
    # where we want to set v itself (path is of length 0).
    for i, el in enumerate(path[:last_settable_index]):
        if v is None:
            return
        if isinstance(el, PathItemStructField):
            v = v[el.key]
            arrow_data = arrow_data.field(el.key)
        elif isinstance(el, PathItemObjectField):
            v = getattr(v, el.attr)
            arrow_data = arrow_data.field(el.attr)
        elif el == PathItemTaggedValueTag():
            try:
                v = v["_tag"]
            except (TypeError, KeyError):
                v = v.tag
            arrow_data = arrow_data.field("_tag")
        elif el == PathItemTaggedValueValue():
            try:
                v = v["_value"]
            except (TypeError, KeyError):
                v = v.value
            arrow_data = arrow_data.field("_value")
        elif el == PathItemOuterList():
            for j, vi in enumerate(v):
                set_path(
                    vi,
                    arrow_data,
                    path[i + 1 :],
                    value_fn,
                    offset + j,
                    arrow_offset + j,
                )
            return
        elif el == PathItemList():
            # its a list
            for j, vi in enumerate(v):
                set_path(
                    vi,
                    arrow_data.values,
                    path[i + 1 :],
                    value_fn,
                    offset + j,
                    arrow_offset + j,
                )
            return
        elif isinstance(el, PathItemUnionEntry):
            if el.type_code != arrow_data.type_codes[arrow_offset].as_py():
                return
            arrow_data = arrow_data.field(el.type_code)
            arrow_offset = 0
        else:
            raise ValueError(f"Unexpected path element: {el}")
    if v is None:
        return

    leaf = path[last_settable_index]

    if isinstance(leaf, PathItemStructField):
        if final_union_path_el:
            arrow_data = arrow_data.field(leaf.key)
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        v[leaf.key] = value_fn(v[leaf.key], offset)
    elif isinstance(leaf, PathItemObjectField):
        if final_union_path_el:
            arrow_data = arrow_data.field(leaf.attr)
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        setattr(v, leaf.attr, value_fn(getattr(v, leaf.attr), offset))
    elif leaf == PathItemTaggedValueTag():
        if final_union_path_el:
            arrow_data = arrow_data.field("_tag")
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        try:
            v["_tag"] = value_fn(v["_tag"], offset)
        except (KeyError, TypeError):
            v.tag = value_fn(v.tag, offset)
    elif leaf == PathItemTaggedValueValue():
        if final_union_path_el:
            arrow_data = arrow_data.field("_value")
            if (
                final_union_path_el.type_code
                != arrow_data.type_codes[arrow_offset].as_py()
            ):
                return
        try:
            v["_value"] = value_fn(v["_value"], offset)
        except (KeyError, TypeError):
            # print("V", v)
            v.value = value_fn(v.value, offset)
    elif leaf == PathItemList():
        if final_union_path_el:
            arrow_data = arrow_data.values
            for j, vi in enumerate(v):
                if (
                    final_union_path_el.type_code
                    != arrow_data.type_codes[arrow_offset + j].as_py()
                ):
                    continue
                v[j] = value_fn(vi, offset + j)
        else:
            for j, vi in enumerate(v):
                v[j] = value_fn(vi, offset + j)
    elif leaf == PathItemOuterList():
        if final_union_path_el:
            for j, vi in enumerate(v):
                if (
                    final_union_path_el.type_code
                    != arrow_data.type_codes[arrow_offset + j].as_py()
                ):
                    continue
                v[j] = value_fn(vi, offset + j)
        else:
            for j, vi in enumerate(v):
                v[j] = value_fn(vi, offset + j)
    else:
        raise ValueError(f"Unexpected path element: {leaf}")


def weave_arrow_type_check(
    wt: types.Type, arr: pa.Array, optional: bool = False
) -> typing.Optional[str]:
    reasons: list[str] = []
    at: pa.DataType = arr.type
    if wt == types.Any():
        reasons.append("Any not allowed")
    elif isinstance(wt, types.UnionType):
        for i in range(len(wt.members)):
            for j in range(i + 1, len(wt.members)):
                if not isinstance(
                    types.merge_types(wt.members[i], wt.members[j]), types.UnionType
                ):
                    reasons.append(f"Union members {i} and {j} are mergable")
        non_none_members = [m for m in wt.members if m != types.NoneType()]
        none_member_count = len(wt.members) - len(non_none_members)
        if none_member_count > 1:
            reasons.append("Multiple None members in Union")
        elif none_member_count == 1:
            # Nullable case
            # TODO: Actually must check nullable!
            reason = None
            if len(non_none_members) == 0:
                reason = "NoneType not allowed in Union with no other members"
            if at != pa.null():
                if len(non_none_members) == 1:
                    reason = weave_arrow_type_check(
                        non_none_members[0], arr, optional=True
                    )
                else:
                    reason = weave_arrow_type_check(
                        types.UnionType(*non_none_members), arr, optional=True
                    )
            if reason is not None:
                reasons.append(f"Nullable member: {reason}")
        else:
            if not pa.types.is_union(at):
                reasons.append(f"Expected UnionType, got {at}")
            else:
                if len(wt.members) != len(at):
                    reasons.append(
                        f"Union length mismatch: {len(wt.members)} != {len(at)}"
                    )
                else:
                    for i, w in enumerate(wt.members):
                        reason = weave_arrow_type_check(
                            w, arr.field(i), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Union member {i}: {reason}")
    elif optional and pa.types.is_null(at):
        pass
    else:
        # Nullable check. This surfaces a ton of failures...
        # if wt != types.NoneType() and not optional and arr.null_count > 0:
        #     reasons.append("Non-nullable but array has nulls")

        if isinstance(wt, tagged_value_type.TaggedValueType):
            tag_reason = weave_arrow_type_check(
                wt.tag, arr.field("_tag"), optional=optional
            )
            if tag_reason is not None:
                reasons.append(f"TaggedValue tag: {tag_reason}")
            value_reason = weave_arrow_type_check(
                wt.value, arr.field("_value"), optional=optional
            )
            if value_reason is not None:
                reasons.append(f"TaggedValue value: {value_reason}")

        elif isinstance(wt, types.Const):
            reasons.append("Const not allowed")

        elif isinstance(wt, types.Function):
            if not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")

        elif isinstance(wt, types.TypedDict):
            if not pa.types.is_struct(at):
                reasons.append(f"Expected StructType, got {at}")
            else:
                at_fields_by_name = {f.name: f for f in at}
                for k, v in wt.property_types.items():
                    if k not in at_fields_by_name:
                        reasons.append(f"Missing field {k}")
                    else:
                        reason = weave_arrow_type_check(
                            v, arr.field(k), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Field {k}: {reason}")

        elif isinstance(wt, (types.List, ArrowWeaveListType)):
            if not pa.types.is_list(at):
                reasons.append(f"Expected ListType, got {at}")
            else:
                reason = weave_arrow_type_check(
                    wt.object_type, arr.flatten(), optional=optional
                )
                if reason is not None:
                    reasons.append(f"List element: {reason}")

        elif isinstance(wt, types.ObjectType):
            if not pa.types.is_struct(at):
                reasons.append(f"Expected StructType, got {at}")
            else:
                at_fields_by_name = {f.name: f for f in at}
                for k, v in wt.property_types().items():
                    if k not in at_fields_by_name:
                        reasons.append(f"Missing field {k}")
                    else:
                        reason = weave_arrow_type_check(
                            v, arr.field(k), optional=optional
                        )
                        if reason is not None:
                            reasons.append(f"Field {k}: {reason}")
        elif isinstance(wt, types.String):
            if pa.types.is_dictionary(at):
                if not pa.types.is_string(at.value_type):
                    reasons.append(
                        f"Expected StringType as dictionary type, got {at.value_type}"
                    )
            elif not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")
        elif isinstance(wt, types.UnknownType):
            if len(arr) != 0:
                reasons.append("Non-zero length array with UnknownType")
        elif isinstance(wt, types.Int):
            # We allow float32/64 because of a bug in our unnest implementation
            # which relies on pandas. TODO: Remove this when we fix unnest.
            if (
                not pa.types.is_float64(at)
                and not pa.types.is_int64(at)
                and not pa.types.is_float32(at)
                and not pa.types.is_int32(at)
            ):
                reasons.append(
                    f"Expected int64 or int32 (float32/64 allowed because of unnest bug), got {at}"
                )
        elif isinstance(wt, types.Float):
            if not pa.types.is_float64(at) and not pa.types.is_float32(at):
                reasons.append(f"Expected float64 or float32, got {at}")
        elif isinstance(wt, types.Number):
            if (
                not pa.types.is_float64(at)
                and not pa.types.is_int64(at)
                and not pa.types.is_float32(at)
                and not pa.types.is_int32(at)
            ):
                reasons.append(
                    f"Expected float64 or float32 or int64 or int32, got {at}"
                )
        elif isinstance(wt, types.Boolean):
            if not pa.types.is_boolean(at):
                reasons.append(f"Expected boolean, got {at}")
        elif isinstance(wt, types.NoneType):
            # We don't check for pa.types.is_null(at) because we can have for
            # example at == int32 with all nulls.
            if not arr.null_count == len(arr):
                reasons.append(f"Expected all nulls, got {arr.null_count}")
        elif isinstance(wt, types.Timestamp):
            if not pa.types.is_timestamp(at):
                reasons.append(f"Expected TimestampType, got {at}")
        elif types.is_custom_type(wt):
            if pa.types.is_dictionary(at):
                if not pa.types.is_string(at.value_type):
                    reasons.append(
                        f"Expected StringType as dictionary type, got {at.value_type}"
                    )
            elif not pa.types.is_string(at):
                reasons.append(f"Expected StringType, got {at}")
        else:
            reasons.append(f"Unhandled case: {wt}, {at}")

    if reasons:
        indented_reasons = textwrap.indent("\n".join(reasons), "  ")
        print("TYPE", type(wt))
        return f"{debug_types.short_type(wt)}, {str(at)[:50]}\n{indented_reasons}"

    return None


_awl_invalid_reason: contextvars.ContextVar[
    typing.Optional[str]
] = contextvars.ContextVar("_awl_invalid_reason", default=None)


@contextlib.contextmanager
def unsafe_awl_construction(reason: str):
    token = None
    if _awl_invalid_reason.get() is None:
        token = _awl_invalid_reason.set(reason)
    try:
        yield
    finally:
        if token is not None:
            _awl_invalid_reason.reset(token)


ArrowWeaveListObjectTypeVar = typing.TypeVar("ArrowWeaveListObjectTypeVar")


def dict_of_columns_to_awl(d: dict[str, typing.Any]) -> "ArrowWeaveList":
    from weave.legacy.weave.arrow import constructors

    cols = {}
    for k, v in d.items():
        if isinstance(v, dict):
            cols[k] = dict_of_columns_to_awl(v)
        elif isinstance(v, list):
            cols[k] = ArrowWeaveList(v)
        else:
            cols[k] = v
    res = constructors.vectorized_container_constructor_preprocessor(cols)
    arrow_data = pa.StructArray.from_arrays(res.arrays, list(cols.keys()))
    object_type = types.TypedDict(res.prop_types)
    return ArrowWeaveList(arrow_data, object_type)


class ArrowWeaveList(typing.Generic[ArrowWeaveListObjectTypeVar]):
    _arrow_data: pa.Array
    object_type: types.Type

    # TODO: Refactor to disable None artifact? (Only used in tests)
    def __init__(
        self,
        arrow_data: typing.Union[pa.Table, pa.ChunkedArray, pa.Array, dict, list],
        object_type=None,
        artifact: typing.Optional[artifact_base.Artifact] = None,
        invalid_reason=None,
    ) -> None:
        from weave.legacy.weave.arrow import constructors, convert

        # Do not dictionary decode this array! That will break performance.
        # Note we combine chunks here, to make the internal interface easy
        # to use. In the future we could refactor to retain chunked form.
        if isinstance(arrow_data, dict):
            awl = dict_of_columns_to_awl(arrow_data)
            self._arrow_data = awl._arrow_data
            object_type = awl.object_type
        elif isinstance(arrow_data, list):
            loaded = convert.to_arrow(arrow_data)
            self._arrow_data = loaded._arrow_data
            object_type = loaded.object_type
        elif isinstance(arrow_data, pa.Array):
            self._arrow_data = arrow_data
        elif isinstance(arrow_data, pa.Table):
            self._arrow_data = pa.StructArray.from_arrays(
                # TODO: we shouldn't need to combine chunks, we can produce this in the
                # original chunked form for zero copy
                [c.combine_chunks() for c in arrow_data.columns],
                names=arrow_data.column_names,
            )
        elif isinstance(arrow_data, pa.ChunkedArray):
            self._arrow_data = arrow_data.combine_chunks()
        else:
            raise TypeError(
                "Expected pyarrow Array, ChunkdArray or Table, got %s"
                % type(arrow_data)
            )

        self.object_type = object_type
        if self.object_type is None:
            self.object_type = types.TypeRegistry.type_of(self._arrow_data).object_type
        self._artifact = artifact
        if invalid_reason is None:
            invalid_reason = _awl_invalid_reason.get()
        self._invalid_reason = invalid_reason

        self._validate()

    @property
    def simple_value_type(self) -> types.Type:
        """Returns the type of values stored in this list, without any transparent types."""
        return types.simple_type(self.object_type)

    def is_number(self) -> bool:
        return isinstance(self.simple_value_type, types.Number)

    def is_boolean(self) -> bool:
        return isinstance(self.simple_value_type, types.Boolean)

    def is_dict(self) -> bool:
        return isinstance(self.simple_value_type, types.TypedDict)

    def _lookup_path(self, path: list[str]):
        assert len(path) > 1
        edge_type = path[0]
        edge_path = path[1]

        # The client might send over refs without knowing the underlying list type is a table.
        if (
            edge_type == ref_util.DICT_KEY_EDGE_NAME
            or edge_type == ref_util.OBJECT_ATTR_EDGE_NAME
        ):
            edge_type = ref_util.AWL_COL_EDGE_NAME
        elif edge_type == ref_util.LIST_INDEX_EDGE_NAME:
            edge_type = ref_util.AWL_ROW_EDGE_NAME

        assert edge_type in [
            ref_util.AWL_ROW_EDGE_NAME,
            ref_util.AWL_COL_EDGE_NAME,
        ]
        if edge_type == ref_util.AWL_ROW_EDGE_NAME:
            res = self[int(edge_path)]
        else:
            res = self.column(edge_path)

        remaining_path = path[2:]
        if remaining_path:
            return res._lookup_path(remaining_path)
        return res

    def __add__(self, other) -> "ArrowWeaveList":
        if isinstance(other, list):
            other = ArrowWeaveList(other)
        if not isinstance(other, ArrowWeaveList):
            raise TypeError(f"Expected list or ArrowWeaveList, got {type(other)}")

        from weave.legacy.weave.arrow import concat

        return concat.concatenate(self, other)

    def _validate(self) -> None:
        if self._invalid_reason is None:
            self._validate_weave_type()
        self._validate_arrow_data()

    def _validate_weave_type(self) -> None:
        return

        # Validation disabled
        arr = self._arrow_data
        type_match_summary = weave_arrow_type_check(self.object_type, arr)
        if type_match_summary != None:
            print()
            print("ArrowWeaveList VALIDATION ERROR")
            print()
            print("ArrowWeaveList.object_type", self.object_type)
            print()
            print(
                "ArrowWeaveList._arrow_data.type\n",
                pretty_print_arrow_type(self._arrow_data.type),
            )
            print()
            print(type_match_summary)
            raise ValueError(f"ArrowWeaveList validation err: {type_match_summary}")

    def _validate_arrow_data(self) -> None:
        return

        # Validation disabled: this seems to usually be fast, but can take
        # many seconds for certain arrays.
        try:
            self._arrow_data.validate(full=True)
        except pa.ArrowInvalid:
            raise

    def validate(self) -> None:
        self._validate_weave_type()
        self._validate_arrow_data()
        self._invalid_reason = None

    def with_column(self, key: str, val: "ArrowWeaveList"):
        """Add a column to the AWL.

        key can be a dot-separated path. Any non-TypedDict values encountered
        in traversal will be replaced with empty TypedDicts.

        leaf will be replaced with val if leaf exists. Added value is always
        added to the end of the TypedDict (remember, they are ordered).

        Result len is always len(self). Will raise if len(val) != len(self)
        """
        if not isinstance(self.object_type, types.TypedDict):
            self = make_vec_dict()
        self_object_type = typing.cast(types.TypedDict, self.object_type)

        path = _dict_utils.split_escaped_string(key)

        if len(path) > 1:
            return self.with_column(
                path[0], self.column(path[0]).with_column(".".join(path[1:]), val)
            )

        col_names = list(self_object_type.property_types.keys())
        col_data = [self._arrow_data.field(i) for i in range(len(col_names))]
        property_types = {**self_object_type.property_types}
        try:
            key_index = col_names.index(key)
        except ValueError:
            key_index = None
        if key_index is not None:
            property_types.pop(key)
            col_names.pop(key_index)
            col_data.pop(key_index)
        col_names.append(key)
        col_data.append(val._arrow_data)
        property_types[key] = val.object_type
        return ArrowWeaveList(
            pa.StructArray.from_arrays(col_data, names=col_names),
            types.TypedDict(property_types),
            self._artifact,
        )

    def with_columns(self, cols: dict[str, "ArrowWeaveList"]):
        for k, v in cols.items():
            self = self.with_column(k, v)
        return self

    def map_column(
        self,
        fn: typing.Callable[
            ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
        ],
        pre_fn: typing.Optional[
            typing.Callable[
                ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
            ]
        ] = None,
    ) -> "ArrowWeaveList":
        return self._map_column(fn, pre_fn, ())

    def _map_column(
        self,
        fn: typing.Callable[
            ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
        ],
        pre_fn: typing.Optional[
            typing.Callable[
                ["ArrowWeaveList", PathType], typing.Optional["ArrowWeaveList"]
            ]
        ],
        path: PathType,
    ) -> "ArrowWeaveList":
        if pre_fn is not None:
            pre_mapped = pre_fn(self, path)
            if pre_mapped is not None:
                self = pre_mapped
        with_mapped_children = self
        if isinstance(self.object_type, types.UnknownType):
            # Only occurs in empty list. Don't map into this.
            return self
        if isinstance(self.object_type, types.Const):
            with_mapped_children = ArrowWeaveList(
                self._arrow_data, self.object_type.val_type, self._artifact
            )._map_column(fn, pre_fn, path)
        if self._arrow_data.type == pa.null():
            # we can have a null array at any type. We stop
            # mapping when we hit one. The caller should specifically handle
            # null arrays in pre or post map functions
            pass
        elif isinstance(self.object_type, types.TypedDict):
            arr = self._arrow_data
            properties: dict[str, ArrowWeaveList] = {
                k: ArrowWeaveList(arr.field(k), v, self._artifact)._map_column(
                    fn, pre_fn, path + (PathItemStructField(k),)
                )
                for k, v in self.object_type.property_types.items()
            }
            if not properties:
                result_arr = arr
            else:
                result_arr = pa.StructArray.from_arrays(
                    [v._arrow_data for v in properties.values()],
                    list(properties.keys()),
                    mask=pa.compute.is_null(arr),
                )

            # set invalid_reason to the first non-None invalid reason found in
            # properties
            invalid_reason = first_non_none(
                v._invalid_reason for v in properties.values()
            )

            with_mapped_children = ArrowWeaveList(
                result_arr,
                types.TypedDict({k: v.object_type for k, v in properties.items()}),
                self._artifact,
                invalid_reason=invalid_reason,
            )
        elif isinstance(self.object_type, types.ObjectType):
            arr = self._arrow_data
            attrs: dict[str, ArrowWeaveList] = {
                k: ArrowWeaveList(arr.field(k), v, self._artifact)._map_column(
                    fn, pre_fn, path + (PathItemObjectField(k),)
                )
                for k, v in self.object_type.property_types().items()
            }

            # Types of some of the attrs may have changed. But some property types on object
            # types are no variable, and it is invalid to change them from our type system's
            # perspective. So the result here is invalid ArrowWeaveList. The results should
            # only be used in cirumstances where this is acceptable.
            with_mapped_children = ArrowWeaveList(
                pa.StructArray.from_arrays(
                    [v._arrow_data for v in attrs.values()],
                    list(attrs.keys()),
                    mask=pa.compute.is_null(arr),
                ),
                self.object_type,
                self._artifact,
                invalid_reason="mapped_column on ObjectType can produce invalid ArrowWeaveList",
            )
        elif isinstance(self.object_type, (types.List, ArrowWeaveListType)):
            arr = self._arrow_data
            items: ArrowWeaveList = ArrowWeaveList(
                arr.flatten(), self.object_type.object_type, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemList(),))
            # print("SELF OBJECT TYPE", self.object_type)
            # print("SELF ARROW DATA TYPE", self._arrow_data.type)
            with_mapped_children = ArrowWeaveList(
                pa.ListArray.from_arrays(
                    offsets_starting_at_zero(self._arrow_data),
                    items._arrow_data,
                    mask=pa.compute.is_null(arr),
                ),
                self.object_type.__class__(items.object_type),
                self._artifact,
                invalid_reason=items._invalid_reason,
            )
        elif isinstance(self.object_type, tagged_value_type.TaggedValueType):
            arr = self._arrow_data
            tag: ArrowWeaveList = ArrowWeaveList(
                self._arrow_data.field("_tag"), self.object_type.tag, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemTaggedValueTag(),))
            value: ArrowWeaveList = ArrowWeaveList(
                self._arrow_data.field("_value"), self.object_type.value, self._artifact
            )._map_column(fn, pre_fn, path + (PathItemTaggedValueValue(),))
            with_mapped_children = ArrowWeaveList(
                pa.StructArray.from_arrays(
                    [tag._arrow_data, value._arrow_data],
                    ["_tag", "_value"],
                    mask=pa.compute.is_null(arr),
                ),
                tagged_value_type.TaggedValueType(tag.object_type, value.object_type),  # type: ignore
                self._artifact,
                invalid_reason=tag._invalid_reason or value._invalid_reason,
            )
        elif isinstance(self.object_type, types.UnionType):
            non_none_members = [
                m for m in self.object_type.members if not isinstance(m, types.NoneType)
            ]
            nullable = len(non_none_members) < len(self.object_type.members)
            if len(non_none_members) == 1:
                non_none_member = ArrowWeaveList(
                    self._arrow_data,
                    non_none_members[0],
                    self._artifact,
                )._map_column(fn, pre_fn, path)
                with_mapped_children = ArrowWeaveList(
                    non_none_member._arrow_data,
                    types.optional(non_none_member.object_type),
                    self._artifact,
                    invalid_reason=non_none_member._invalid_reason,
                )
            elif len(non_none_members) > 1:
                arr = self._arrow_data
                members: list[typing.Optional[ArrowWeaveList]] = [
                    ArrowWeaveList(
                        arr.field(i),
                        types.optional(member_type) if nullable else member_type,
                        self._artifact,
                    )._map_column(fn, pre_fn, path + (PathItemUnionEntry(i),))
                    for i, member_type in enumerate(non_none_members)
                ]

                # Types of some of the members may have changed. We need to maintain
                # the invariant that types within a union are not mergeable via
                # merge_types. So we walk through the new members looking for
                # mergeable types, and concatenating them, while computing
                # new type_codes and offsets arrays

                type_codes = self._arrow_data.type_codes
                offsets = self._arrow_data.offsets
                for i in range(len(members)):
                    if members[i] is None:
                        # We've already merged this member into another, so
                        # find the first non-None member to start from.
                        for j in range(i + 1, len(members)):
                            if members[j] is not None:
                                members[i] = members[j]
                                members[j] = None
                                type_codes = pc.choose(
                                    pc.equal(type_codes, j),
                                    type_codes,
                                    pa.scalar(i, pa.int8()),
                                )
                                break
                    for j in range(i + 1, len(members)):
                        member_i = members[i]
                        member_j = members[j]
                        if (
                            member_i is not None
                            and member_j is not None
                            and types.types_are_mergeable(
                                member_i.object_type, member_j.object_type
                            )
                        ):
                            merged = True
                            offsets = pc.choose(
                                pc.equal(type_codes, j),
                                offsets,
                                pc.add(offsets, len(member_i)),
                            ).cast(pa.int32())
                            type_codes = pc.choose(
                                pc.equal(type_codes, j),
                                type_codes,
                                pa.scalar(i, pa.int8()),
                            )
                            from weave.legacy.weave.arrow import concat

                            members[i] = concat.concatenate(member_i, member_j)
                            members[j] = None

                final_members: list[ArrowWeaveList] = [
                    m for m in members if m is not None
                ]
                if len(final_members) == 1:
                    with_mapped_children = ArrowWeaveList(
                        final_members[0]._arrow_data.take(offsets),
                        final_members[0].object_type,
                        final_members[0]._artifact,
                    )
                else:
                    new_type_members = [m.object_type for m in final_members]
                    if nullable:
                        new_type_members.append(types.NoneType())

                    invalid_reason = first_non_none(
                        m._invalid_reason for m in final_members
                    )
                    with_mapped_children = ArrowWeaveList(
                        pa.UnionArray.from_dense(
                            type_codes,
                            offsets,
                            [m._arrow_data for m in final_members],
                        ),
                        types.UnionType(*new_type_members),
                        # types.union(*new_type_members),
                        self._artifact,
                        invalid_reason=invalid_reason,
                    )
            else:
                raise errors.WeaveInternalError(
                    "Union must have at least one non-none member"
                )

        mapped = fn(with_mapped_children, path)
        if mapped is None:
            mapped = with_mapped_children
        return mapped

    def separate_tags(
        self,
    ) -> typing.Tuple["ArrowWeaveList", dict[PathType, "ArrowWeaveList"]]:
        tag_columns: dict[PathType, ArrowWeaveList] = {}

        def _remove_tags(list: ArrowWeaveList, path: PathType) -> ArrowWeaveList:
            if isinstance(list.object_type, tagged_value_type.TaggedValueType):
                # We need to remove _value components in the case that we are inside
                # tagged value here.
                tag_columns[path] = ArrowWeaveList(
                    list._arrow_data.field("_tag"), list.object_type.tag, list._artifact
                )
                return ArrowWeaveList(
                    list._arrow_data.field("_value"),
                    list.object_type.value,
                    list._artifact,
                )
            return list

        res = self.map_column(_remove_tags)

        reversed_tag_columns = {}
        for k, v in reversed(tag_columns.items()):
            reversed_tag_columns[k] = v

        return res, reversed_tag_columns

    def without_tags(self) -> "ArrowWeaveList":
        return self.separate_tags()[0]

    def separate_dictionaries(
        self,
    ) -> typing.Tuple["ArrowWeaveList", dict[PathType, "ArrowWeaveList"]]:
        dictionary_columns: dict[PathType, ArrowWeaveList] = {}

        def _remove_dictionaries(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if pa.types.is_dictionary(list._arrow_data.type):
                dictionary_columns[path] = ArrowWeaveList(
                    list._arrow_data.dictionary, list.object_type, list._artifact
                )
                return ArrowWeaveList(
                    list._arrow_data.indices,
                    # Don't change the object type to Int. Instead, we mark the ArrowWeaveList
                    # as invalid (the arrow type no longer matches object_type)
                    list.object_type,
                    list._artifact,
                    invalid_reason="Dictionary removed",
                )
            return None

        res = self.map_column(_remove_dictionaries), dictionary_columns
        return res

    def without_dictionaries(self) -> "ArrowWeaveList":
        def _remove_dictionaries(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if pa.types.is_dictionary(list._arrow_data.type):
                return ArrowWeaveList(
                    list._arrow_data.dictionary_decode(),
                    list.object_type,
                    list._artifact,
                    invalid_reason="Dictionary removed",
                )
            return None

        return self.map_column(_remove_dictionaries)

    def separate_awls(
        self,
    ) -> typing.Tuple[
        "ArrowWeaveList", dict[PathType, list[typing.Optional["ArrowWeaveList"]]]
    ]:
        awl_columns: dict[PathType, list[typing.Optional[ArrowWeaveList]]] = {}

        def _remove_awls(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if isinstance(list.object_type, ArrowWeaveListType):
                awl_columns[path] = [
                    ArrowWeaveList(
                        a.values, list.object_type.object_type, list._artifact
                    )
                    if a.values is not None
                    else None
                    for a in list._arrow_data
                ]
                return ArrowWeaveList(
                    pa.array([None] * len(list)), types.NoneType(), list._artifact
                )
            return None

        # Use pre-order traversal. We only want the top most awls.
        return self.map_column(lambda x, y: None, _remove_awls), awl_columns

    def function_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        function_type_paths: dict[PathType, types.Type] = {}

        def _save_obj_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(list.object_type, types.Function):
                function_type_paths[path] = list.object_type
            return None

        self.map_column(_save_obj_type_path)
        function_type_paths = reverse_dict(function_type_paths)
        return function_type_paths

    def obj_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        obj_type_paths: dict[PathType, types.Type] = {}

        def _save_obj_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(list.object_type, types.ObjectType):
                obj_type_paths[path] = list.object_type
            return None

        self.map_column(_save_obj_type_path)
        obj_type_paths = reverse_dict(obj_type_paths)
        return obj_type_paths

    def custom_type_paths(
        self,
    ) -> dict[PathType, types.Type]:
        custom_type_paths: dict[PathType, types.Type] = {}

        def _save_custom_type_path(list: ArrowWeaveList, path: PathType) -> None:
            if isinstance(
                list.object_type,
                # TODO: don't hardcode
                (
                    ArrowWeaveListType,
                    types.List,
                    types.Const,
                    types.Number,
                    types.Int,
                    types.Float,
                    types.Boolean,
                    types.String,
                    types.Timestamp,
                    types.NoneType,
                    tagged_value_type.TaggedValueType,
                    types.TypedDict,
                    types.ObjectType,
                    types.UnionType,
                    types.Function,
                    types.TimeDelta,
                ),
            ):
                return None
            custom_type_paths[path] = list.object_type

        self.map_column(_save_custom_type_path)
        return custom_type_paths

    def _arrow_data_asarray_no_tags(self) -> pa.Array:
        # Remove both tags and dictionaries. You should typically leave
        # dictionaries as late as possible! So only use this if you're sure
        # you don't want them.
        no_tags = self.without_tags()
        no_dicts = no_tags.without_dictionaries()
        return no_dicts._arrow_data

    def __array__(self, dtype=None):
        # TODO: replace with to_pylist_tagged once refs are supported
        pylist = self.to_pylist_raw()
        return np.asarray(pylist)

    def __iter__(self):
        self_ref = ref_base.get_ref(self)
        pylist = self.to_pylist_tagged()
        for i, x in enumerate(pylist):
            yield ref_util.val_with_relative_ref(
                self, x, [ref_util.AWL_ROW_EDGE_NAME, str(i)]
            )

    def __repr__(self):
        return f"<ArrowWeaveList: {self.object_type}>"

    def to_pylist_raw(self):
        """Used for testing, preserves _tag and _value fields"""
        return self._to_pylist_dictsafe()

    def _to_pylist_dictsafe(self):
        value_awl, dict_columns = self.separate_dictionaries()
        value_py = value_awl._arrow_data.to_pylist()

        dict_columns = {p: c._arrow_data.to_pylist() for p, c in dict_columns.items()}

        # Dictionary decode the value
        if dict_columns:
            for path, dict_col in dict_columns.items():
                path = tuple(
                    p
                    if not isinstance(p, PathItemObjectField)
                    else PathItemStructField(p.attr)
                    for p in path
                )
                set_path(
                    value_py,
                    value_awl._arrow_data,
                    (PathItemOuterList(),) + path,
                    lambda v, j: None if v == None else dict_col[v],
                )

        return value_py

    def to_pylist_notags(self):
        """Convert the ArrowWeaveList to a python list, stripping tags"""
        value_awl, _ = self.separate_tags()
        return value_awl._to_pylist_dictsafe()

    def to_pylist_tagged(self):
        """Convert the ArrowWeaveList to a python list, tagging objects correctly"""
        value, awl_columns = self.separate_awls()

        custom_type_paths = value.custom_type_paths()
        function_type_paths = value.function_type_paths()
        obj_type_paths = value.obj_type_paths()

        value_awl, dict_columns = value.separate_dictionaries()
        value_py = value_awl._arrow_data.to_pylist()

        dict_columns = {p: c._arrow_data.to_pylist() for p, c in dict_columns.items()}

        for path, obj_type in obj_type_paths.items():
            non_op_def_keys = [
                k
                for k, t in obj_type.property_types().items()
                if not t == op_def_type.OpDefType()
            ]

            def val_to_obj(v, y):
                # We init the object with non-op-def (non-method) keys.
                # This is not quite right, this code needs to do what we do
                # in mappers_python, and also attach the methods to the object
                # after construction.

                # So we don't yet correctly deserialize lists of relocatable objects
                # with different methods correctly.
                # TODO: fix
                return (
                    None
                    if v is None
                    else obj_type.instance_class(**{k: v[k] for k in non_op_def_keys})
                )

            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                val_to_obj,
            )
        # Dictionary decode the value, and add the tags to the tag store,
        # in a single pass.
        for path2, dict_col in dict_columns.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path2,
                lambda v, j: None if v == None else dict_col[v],
            )

        for path, type_ in custom_type_paths.items():

            def custom_type_value(v, j):
                if v == None:
                    return None
                if isinstance(v, dict):
                    return type_.instance_from_dict(v)
                if isinstance(v, op_def.OpDef):
                    # I'm not sure how this is already an OpDef, but it is
                    # so we can just return...
                    return v
                # else its a ref string
                # TODO: this does not use self.artifact, can we just drop it?
                # Do we need the type so we can load here? No...
                if ":" in v:
                    ref = ref_base.Ref.from_str(v)
                    # Note: we are forcing type here, because we already know it
                    # We don't save the types for every file in a remote artifact!
                    # But you can still reference them, because you have to get that
                    # file through an op, and therefore we know the type?
                    ref._type = type_

                    if isinstance(type_, types.RefType):
                        ref._type = type_.object_type
                        # if we actually have a ref, then don't deref, return the ref
                        return ref

                    return ref.get()

                return self._artifact.get(v, type_)

            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                custom_type_value,
            )

        for path, function_type in function_type_paths.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                lambda v, j: node_ref.ref_to_node(ref_base.Ref.from_str(v)),
            )

        for path, awls in awl_columns.items():
            set_path(
                value_py,
                value_awl._arrow_data,
                (PathItemOuterList(),) + path,
                lambda v, j: awls[j],
            )
        return concrete_to_tagstore(value_py)

    def _count(self):
        return len(self._arrow_data)

    def __len__(self):
        return self._count()

    def replace_column(
        self,
        name: typing.Union[str, PathType],
        fn: typing.Callable[["ArrowWeaveList"], "ArrowWeaveList"],
    ):
        replaced = {"v": False}

        def _do_replace(
            list: ArrowWeaveList, path: PathType
        ) -> typing.Optional[ArrowWeaveList]:
            if (
                isinstance(name, tuple)
                and path == name
                or len(path) == 1
                and path[0] == name
            ):
                replaced["v"] = True
                return fn(list)
            return None

        res = self.map_column(_do_replace)
        if not replaced["v"]:
            raise ValueError("Column {} not found".format(name))
        return res

    def dictionary_encode(self) -> "ArrowWeaveList":
        return ArrowWeaveList(
            self._arrow_data.dictionary_encode(), self.object_type, self._artifact
        )

    def _index(
        self,
        index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
    ):
        if index == None:
            return None
        indexes: pa.Array
        if isinstance(index, int):
            indexes = [index]
        elif isinstance(index, ArrowWeaveList):
            indexes = index._arrow_data
        else:
            indexes = index

        arr = self._arrow_data
        length = len(arr)
        neg_cond = pc.less(indexes, 0)
        indexes_neg = pc.add(indexes, length)
        indexes = pc.choose(neg_cond, indexes, indexes_neg)
        out_of_bounds = pc.or_(pc.less(indexes, 0), pc.greater_equal(indexes, length))
        indexes_bounds_checked = pc.choose(
            out_of_bounds, indexes, pa.scalar(None, pa.int64())
        )
        result_rows = pc.take(arr, indexes_bounds_checked)
        awl: ArrowWeaveList = ArrowWeaveList(
            result_rows, self.object_type, self._artifact
        )
        if isinstance(index, int):
            result = awl.to_pylist_tagged()[0]

            return ref_util.val_with_relative_ref(
                self, result, [ref_util.AWL_ROW_EDGE_NAME, str(index)]
            )
        return awl

    def __getitem__(
        self,
        index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
    ):
        return self._index(index)

    def _slice(self, start: int, stop: int):
        # We don't slice, we take to get a copy. Arrow slicing is confusing.
        # I was seeing that type_codes in sliced unions were offset somehow...
        return ArrowWeaveList(
            self._arrow_data.take(np.arange(start, stop)),
            self.object_type,
            self._artifact,
        )

    def _append_column(self, name: str, data) -> "ArrowWeaveList":
        if not data:
            raise ValueError(f'Data for new column "{name}" must be nonnull.')

        if isinstance(self._arrow_data, pa.Table):
            new_data = self._arrow_data.append_column(name, [data])
        elif isinstance(self._arrow_data, pa.StructArray):
            chunked_arrays = {}
            for field in self._arrow_data.type:
                chunked_arrays[field.name] = pa.chunked_array(
                    self._arrow_data.field(field.name)
                )
            arrow_obj = pa.table(chunked_arrays)
            new_data = arrow_obj.append_column(name, [data])
        else:
            raise ValueError(
                f"Cannot append column to {type(self._arrow_data)} object."
            )

        return ArrowWeaveList(new_data, None, self._artifact)

    def keys(self) -> list[str]:
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            value_type = self.object_type.value
            if isinstance(value_type, types.TypedDict):
                return list(value_type.property_types.keys())
        elif isinstance(self.object_type, types.TypedDict):
            return list(self.object_type.property_types.keys())
        raise ValueError("Cannot get keys from non-TypedDict.")

    def _split_none(self) -> typing.Tuple[bool, "ArrowWeaveList"]:
        if isinstance(self.object_type, types.UnionType):
            optional, non_none_type = types.split_none(self.object_type)
            return optional, ArrowWeaveList(
                self._arrow_data,
                non_none_type,
                self._artifact,
                invalid_reason="maybe optional",
            )
        return False, self

    def _as_optional(self) -> "ArrowWeaveList":
        invalid_reason = self._invalid_reason
        if self._invalid_reason == "maybe optional":
            invalid_reason = None
        return ArrowWeaveList(
            self._arrow_data,
            types.optional(self.object_type),
            self._artifact,
            invalid_reason=invalid_reason,
        )

    def _null_mask(self) -> pa.Array:
        return safe_is_null(self._arrow_data)

    @property
    def column_names(self):
        obj_type = types.unwrap_type(self.object_type)
        return obj_type.property_types.keys()

    def column(self, name: str) -> "ArrowWeaveList":
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            return make_vec_taggedvalue(
                self.tagged_value_tag(),
                self.tagged_value_value().column(name),
                is_null_mask=self._null_mask(),
            )
        elif isinstance(self.object_type, types.UnionType):
            optional, non_none = self._split_none()
            if optional and not isinstance(non_none.object_type, types.UnionType):
                return non_none.column(name)._as_optional()
            raise ValueError(
                f"Cannot get column {name} from non-TypedDict or ObjectType."
            )
        elif isinstance(self.object_type, types.TypedDict):
            property_types = self.object_type.property_types
        elif isinstance(self.object_type, types.ObjectType):
            property_types = self.object_type.property_types()
        else:
            raise ValueError(
                f"Cannot get column {name} from non-TypedDict or ObjectType."
            )

        if name not in property_types:
            return make_vec_none(len(self))

        val: ArrowWeaveList = ArrowWeaveList(
            self._arrow_data.field(name),
            property_types[name],
            self._artifact,
            invalid_reason=self._invalid_reason,
        )

        return ref_util.val_with_relative_ref(
            self, val, [ref_util.AWL_COL_EDGE_NAME, str(name)]
        )

    def unique(self) -> "ArrowWeaveList":
        return ArrowWeaveList(
            pa.compute.unique(self._arrow_data),
            types.split_none(self.object_type)[1],
            self._artifact,
        )

    def tagged_value_tag(self) -> "ArrowWeaveListGeneric[types.TypedDict]":
        if not isinstance(self.object_type, tagged_value_type.TaggedValueType):
            raise ValueError("Cannot get tagged_value_tag from non-TaggedValueType")

        return ArrowWeaveListGeneric[types.TypedDict](
            self._arrow_data.field("_tag"), self.object_type.tag, self._artifact
        )

    def tagged_value_value(self) -> "ArrowWeaveList":
        if not isinstance(self.object_type, tagged_value_type.TaggedValueType):
            raise ValueError("Cannot get tagged_value_tag from non-TaggedValueType")

        return ArrowWeaveList(
            self._arrow_data.field("_value"), self.object_type.value, self._artifact
        )

    def untagged(self) -> "ArrowWeaveList":
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            return self.tagged_value_value()
        return self

    def _make_lambda_node(
        self, fn: typing.Union[typing.Callable[[typing.Any], typing.Any], graph.Node]
    ):
        if isinstance(fn, graph.Node):
            return fn

        sig = inspect.signature(fn)
        if len(sig.parameters) == 1:
            vars = {"row": self.object_type}
        elif len(sig.parameters) == 2:
            vars = {"row": self.object_type, "index": types.Int()}
        else:
            raise ValueError(
                "Functions passed to ArrowWeaveList.apply must have 1 or 2 parameters (row, [index])"
            )
        return weave_internal.define_fn(vars, fn).val

    # TODO: This should be .map!
    def apply(
        self, fn: typing.Union[typing.Callable[[typing.Any], typing.Any], graph.Node]
    ):
        fn = self._make_lambda_node(fn)
        from weave.legacy.weave.ops_arrow.vectorize import _apply_fn_node_with_tag_pushdown

        return _apply_fn_node_with_tag_pushdown(self, fn)  # type: ignore

    def concat(self, other: "ArrowWeaveList") -> "ArrowWeaveList":
        from weave.legacy.weave.arrow import concat

        return concat.concatenate(self, other)

    def join2(
        self: "ArrowWeaveList",
        other: "ArrowWeaveList",
        join1Fn: graph.OutputNode,
        join2Fn: graph.OutputNode,
        alias1: typing.Optional[str] = None,
        alias2: typing.Optional[str] = None,
        leftOuter: bool = False,
        rightOuter: bool = False,
    ):
        from weave.legacy.weave.ops_arrow import list_join

        join1Fn = self._make_lambda_node(join1Fn)
        join2Fn = other._make_lambda_node(join2Fn)

        return list_join.join2_impl(
            self, other, join1Fn, join2Fn, alias1, alias2, leftOuter, rightOuter
        )

    def _limit(self, limit: int):
        return ArrowWeaveList(
            self._arrow_data.slice(0, limit), self.object_type, self._artifact
        )

    def _with_object_type(self, object_type: types.Type, invalid_reason=None):
        return ArrowWeaveList(
            self._arrow_data, object_type, self._artifact, invalid_reason=invalid_reason
        )

    def _clear_invalid_reason(self):
        return ArrowWeaveList(
            self._arrow_data, self.object_type, self._artifact, invalid_reason=None
        )

    def to_pandas(self):
        arrow_data = self._arrow_data
        if isinstance(self.object_type, types.TypedDict):
            arrow_data = pa.Table.from_arrays(
                [arrow_data.field(i) for i in range(len(arrow_data.type))],
                names=[f.name for f in arrow_data.type],
            )
        return arrow_data.to_pandas()


ArrowWeaveListGenericType = typing.TypeVar(
    "ArrowWeaveListGenericType", bound=types.Type
)


class ArrowWeaveListGeneric(
    ArrowWeaveList[typing.Any], typing.Generic[ArrowWeaveListGenericType]
):
    object_type: ArrowWeaveListGenericType


ArrowWeaveListType.instance_classes = ArrowWeaveList


ArrowWeaveListType.instance_classes = ArrowWeaveList
ArrowWeaveListType.instance_class = ArrowWeaveList


def is_typedict_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.TypedDict]]:
    return isinstance(val.object_type, types.TypedDict)


def is_object_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.ObjectType]]:
    return isinstance(val.object_type, types.ObjectType)


def is_taggedvalue_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[
    ArrowWeaveListGeneric[tagged_value_type.TaggedValueType]
]:
    return isinstance(val.object_type, tagged_value_type.TaggedValueType)


def is_list_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.List]]:
    # Don't use assignability to List here! Because Ref<List> is assignable to List
    return isinstance(val.object_type, types.List) or isinstance(
        val.object_type, ArrowWeaveListType
    )


def is_ref_arrowweavelist(
    val: ArrowWeaveList,
) -> typing_extensions.TypeGuard[ArrowWeaveListGeneric[types.RefType]]:
    return isinstance(val.object_type, types.RefType)


def dataframe_to_arrow(df):
    return ArrowWeaveList(pa.Table.from_pandas(df))


pandas_to_awl = dataframe_to_arrow


def make_vec_none(length: int) -> ArrowWeaveList:
    return ArrowWeaveList(pa.nulls(length), types.NoneType(), None)


def make_vec_dict(**kwargs: ArrowWeaveList):
    arr = pa.StructArray.from_arrays(
        [v._arrow_data for v in kwargs.values()], [k for k in kwargs.keys()]
    )
    property_types = {k: v.object_type for k, v in kwargs.items()}
    return ArrowWeaveList(arr, types.TypedDict(property_types), None)


def make_vec_taggedvalue(
    tag: ArrowWeaveListGeneric[types.TypedDict],
    value: ArrowWeaveList,
    is_null_mask: typing.Optional[pa.Array] = None,
):
    tag_types = tag.object_type.property_types
    value_type = value.object_type
    tag_data = tag._arrow_data
    value_data = value._arrow_data
    if isinstance(value_type, tagged_value_type.TaggedValueType):
        # If value is a tagged value, we flatten by merging the tags and
        # pulling the value up.
        value_type_types = value_type.tag.property_types
        new_tag_types = {**tag_types, **value_type_types}
        if len(new_tag_types) != len(tag_types) + len(value_type_types):
            raise ValueError("Tagged value types must be disjoint")
        tag_types = new_tag_types
        value_type = value_type.value

        value_tag_data = value_data.field("_tag")
        tag_data = pa.StructArray.from_arrays(
            [tag_data.field(f.name) for f in tag_data.type]
            + [value_tag_data.field(f.name) for f in value_tag_data.type],
            names=[f.name for f in tag_data.type]
            + [f.name for f in value_tag_data.type],
        )
        value_data = value_data.field("_value")
    arr = pa.StructArray.from_arrays(
        [tag_data, value_data],
        ["_tag", "_value"],
        mask=is_null_mask,
    )
    return ArrowWeaveList(
        arr,
        tagged_value_type.TaggedValueType(types.TypedDict(tag_types), value_type),
        None,  # type: ignore
    )


def awl_zip(*arrs: ArrowWeaveList) -> ArrowWeaveList:
    if not arrs:
        raise ValueError("Cannot zip empty list")
    from weave.legacy.weave.arrow import convert

    arrs = convert.unify_types(*arrs)
    zipped = arrow_zip(*[a._arrow_data for a in arrs])
    return ArrowWeaveList(zipped, types.List(arrs[0].object_type), None)


# Converts timestamp columns to ms for serialization back to web client
def convert_arrow_timestamp_to_epoch_ms(awl: ArrowWeaveList):
    def convert(col: ArrowWeaveList, path: PathType) -> typing.Optional[ArrowWeaveList]:
        _, non_none_type = types.split_none(col.object_type)
        if isinstance(non_none_type, types.Timestamp):
            return ArrowWeaveList(
                pc.milliseconds_between(
                    pa.scalar(0, col._arrow_data.type), col._arrow_data
                ),
                types.optional(types.Int()),
                awl._artifact,
            )
        return None

    return awl.map_column(convert)



import os
import time

import pytest

from weave.legacy.weave import api as weave
from weave.legacy.weave import (
    box,
    context,
    context_state,
    graph,
    ops,
    storage,
    weave_internal,
)
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_domain import table as table_ops

from weave.legacy.tests.util import weavejs_ops

TABLE_TYPES = ["list", "pandas", "sql"]

_loading_builtins_token = context_state.set_loading_built_ins()


@weave.op(
    name="test_table_ops-op_list_table",
    input_type={"n": types.Int()},
    output_type=types.List(types.TypedDict({"a": types.Int(), "b": types.String()})),
)
def op_list_table(n):
    return [{"a": i, "b": str(i)} for i in range(n)]


@weave.op(
    name="test_table_ops-op_list",
    input_type={"n": types.Int()},
    output_type=types.List(types.TypedDict({"a": types.Int(), "b": types.String()})),
)
def op_list(n):
    return [{"a": i, "b": str(i)} for i in range(n)]


context_state.clear_loading_built_ins(_loading_builtins_token)


def get_test_table(table_type):
    if table_type == "list":
        f = ops.local_path(os.path.join("testdata", "cereal.csv"))
        return f.readcsv()
    elif table_type == "pandas":
        f = ops.local_path(os.path.join("testdata", "cereal.csv"))
        return ops.pandasreadcsv(f)
    elif table_type == "sql":
        c = ops.local_sqlconnection("sqlite:///testdata/cereal.db")
        return ops.sqlconnection_table(c, "cereal")


@pytest.mark.parametrize("table_type", TABLE_TYPES)
def test_count(table_type):
    table = get_test_table(table_type)
    expected = 77
    assert weave.use(table.count()) == expected
    assert weave.use(weavejs_ops.count(table)) == expected


@pytest.mark.parametrize("table_type", TABLE_TYPES)
def test_index(table_type):
    table = get_test_table(table_type)
    expected = {
        "name": "100% Bran",
        "mfr": "N",
        "type": "C",
        "calories": 70,
        "protein": 4,
        "fat": 1,
        "sodium": 130,
        "fiber": 10.0,
        "carbo": 5.0,
        "sugars": 6,
        "potass": 280,
        "vitamins": 25,
        "shelf": 3,
        "weight": 1.0,
        "cups": 0.33,
        "rating": 68.402973,
    }
    assert weave.use(table[0]) == expected
    assert weave.use(weavejs_ops.index(table, 0)) == expected


def js_op_mapped_pick(obj, key):
    return weave_internal.make_output_node(
        types.List(obj.type.object_type.property_types[key]),
        "pick",
        {"obj": obj, "key": weave_internal.make_const_node(weave.types.String(), key)},
    )


@pytest.mark.parametrize("table_type", TABLE_TYPES)
def test_pick(table_type):
    table = get_test_table(table_type)
    assert len(weave.use(table.pick("type"))) == 77
    assert len(weave.use(js_op_mapped_pick(table, "type"))) == 77


@pytest.mark.parametrize("table_type", TABLE_TYPES)
def test_filter(table_type):
    table = get_test_table(table_type)
    # Use the lambda passing convention here.
    filter_node = table.filter(lambda row: row["potass"] > 280)
    assert weave.use(filter_node.count()) == 2
    node = weavejs_ops.filter(table, filter_node.from_op.inputs["filterFn"]).count()
    assert weave.use(node) == 2


# WARNING: Separating tests for group by, because
#   pandas removes the group key sometimes, while the
#   other flavors do not.
@pytest.mark.parametrize("table_type", ["pandas"])
def test_groupby(table_type):
    table = get_test_table(table_type)
    groupby_fn = weave_internal.define_fn(
        {"row": weave.types.TypedDict({})}, lambda row: row["type"]
    )
    # TODO: add a pick here to check that it works.
    # TODO: add a pick test for the array case
    # TODO: add some kind of test that relies on type refinement
    expected = [
        "C",
        {
            "name": "100% Bran",
            "mfr": "N",
            "calories": 70,
            "protein": 4,
            "fat": 1,
            "sodium": 130,
            "fiber": 10.0,
            "carbo": 5.0,
            "sugars": 6,
            "potass": 280,
            "vitamins": 25,
            "shelf": 3,
            "weight": 1.0,
            "cups": 0.33,
            "rating": 68.402973,
        },
    ]
    grouped = table.groupby(groupby_fn)
    assert weave.use((grouped[0].groupkey(), grouped[0][0])) == expected


@pytest.mark.parametrize("table_type", ["list", "sql"])
def test_groupby_list(table_type):
    table = get_test_table(table_type)
    grouped = table.groupby(lambda row: row["type"])
    group0 = grouped[0]
    group0key = group0.groupkey()
    group00 = group0[0]
    expected = [
        "C",
        {
            "name": "100% Bran",
            "mfr": "N",
            "calories": 70,
            "protein": 4,
            "fat": 1,
            "sodium": 130,
            "fiber": 10.0,
            "carbo": 5.0,
            "sugars": 6,
            "type": "C",
            "potass": 280,
            "vitamins": 25,
            "shelf": 3,
            "weight": 1.0,
            "cups": 0.33,
            "rating": 68.402973,
        },
    ]
    assert weave.use((group0key, group00)) == expected


@pytest.mark.parametrize("table_type", ["list", "sql"])
def test_groupby_list_weavejs_form(table_type):
    table = get_test_table(table_type)
    groupby_fn = weave_internal.define_fn(
        {"row": table.type.object_type},
        lambda row: graph.OutputNode(
            types.String(),
            "pick",
            {"obj": row, "key": graph.ConstNode(types.String(), "type")},
        ),
    )
    grouped = weavejs_ops.groupby(table, groupby_fn)
    gr0 = weavejs_ops.index(grouped, 0)
    gr00 = weavejs_ops.index(gr0, 0)
    gr00name = weavejs_ops.weavejs_pick(gr00, "name")
    assert weave.use(gr00name) == "100% Bran"


@pytest.mark.parametrize("table_type", TABLE_TYPES)
def test_map(table_type):
    table = get_test_table(table_type)
    map_fn = weave_internal.define_fn(
        {"row": weave.types.TypedDict({})}, lambda row: row["potass"]
    )
    mapped = table.map(map_fn)
    assert weave.use(mapped[0]) == 280


def test_list_returning_op():
    res = weave.use(op_list_table(2))
    expected = [{"a": 0, "b": str(0)}, {"a": 1, "b": str(1)}]
    assert res == expected
    saved = storage.save(res)
    loaded = storage.deref(saved)
    py = storage.to_python(loaded)["_val"]
    assert py == expected


def test_list_map():
    map_fn = weave_internal.define_fn(
        {"row": weave.types.TypedDict({})}, lambda row: row["a"]
    )
    res = weave.use(op_list_table(2).map(map_fn))
    assert res == [0, 1]


def test_list_get_and_op():
    l = weave.use(op_list(2))
    saved = storage.save(l)
    get_node = ops.get(str(saved))

    # The frontend always sends ops.Table.count() (not the same as get_node.count() right
    # now!)
    count_node = weavejs_ops.count(get_node)
    assert weave.use(count_node) == 2


def test_list_save_and_use():
    saved = storage.save([{"a": 5, "b": 6}], "test-list")
    get_node = ops.get(str(saved))
    with context.weavejs_client():
        assert weave.use(get_node) == [{"a": 5, "b": 6}]


def test_groupby_pick():
    items = weave.save([{"a": 5, "b": 6}, {"a": 5, "b": 8}])
    picked = items.groupby(lambda row: row["a"])[0].pick("b")
    assert weave.use(picked.sum() == 14)


def test_rows_type_null():
    type_node = table_ops.Table.rows_type(None)
    assert weave.use(type_node) == types.NoneType()

    type_node = table_ops.Table.rows_type(box.box(None))
    assert weave.use(type_node) == types.NoneType()


def test_sort_timestamp():
    data = [time.time() * 1000 + x for x in range(10)]
    times = weave.save(data)
    timestamps = times.toTimestamp()
    sorted_ts = timestamps.sort(lambda ts: ops.make_list(label=ts), ["asc"])
    assert weave.use(sorted_ts)[0] == weave.use(timestamps[0])

    sorted_ts = sorted_ts.sort(lambda ts: ops.make_list(label=ts), ["desc"])
    assert weave.use(sorted_ts)[0] == weave.use(timestamps[-1])



import typing

from weave.legacy.weave import urls
from weave.trace.pypi_version_check import check_available

if typing.TYPE_CHECKING:
    import packaging.version  # type: ignore[import-not-found]

REQUIRED_WANDB_VERSION = "0.16.4"


def _parse_version(version: str) -> "packaging.version.Version":
    """Parse a version string into a version object.

    This function is a wrapper around the `packaging.version.parse` function, which
    is used to parse version strings into version objects. If the `packaging` library
    is not installed, it falls back to the `pkg_resources` library.
    """
    try:
        from packaging.version import parse as parse_version  # type: ignore
    except ImportError:
        from pkg_resources import parse_version

    return parse_version(version)


def _print_version_check() -> None:
    import wandb

    import weave

    if _parse_version(REQUIRED_WANDB_VERSION) > _parse_version(wandb.__version__):
        message = (
            "wandb version >= 0.16.4 is required.  To upgrade, please run:\n"
            " $ pip install wandb --upgrade"
        )
        print(message)
    else:
        wandb_messages = check_available(wandb.__version__, "wandb")
        if wandb_messages:
            # Don't print the upgrade message, only the delete or yank message
            use_message = wandb_messages.get("delete_message") or wandb_messages.get(
                "yank_message"
            )  #  or wandb_messages.get("upgrade_message")
            if use_message:
                print(use_message)

    weave_messages = check_available(weave.__version__, "weave")
    if weave_messages:
        use_message = (
            weave_messages.get("delete_message")
            or weave_messages.get("yank_message")
            or weave_messages.get("upgrade_message")
        )
        if use_message:
            print(use_message)


def assert_min_weave_version(
    min_required_version: str, trace_server_host: str = "https://trace.wandb.ai"
) -> None:
    import weave

    if _parse_version(min_required_version) > _parse_version(weave.__version__):
        message = (
            f"The target Weave host {trace_server_host} requires a `weave` package version >= {min_required_version}."
            " To resolve, either:\n"
            "   * Upgrade `weave` by running: `pip install weave --upgrade`.\n"
            "   * Disable logging by omitting calls to `weave.init`."
        )
        raise ValueError(message)


def print_init_message(
    username: typing.Optional[str],
    entity_name: str,
    project_name: str,
    read_only: bool,
) -> None:
    try:
        _print_version_check()
    except Exception as e:
        pass

    message = ""
    if username is not None:
        message += f"Logged in as Weights & Biases user: {username}.\n"
    message += (
        f"View Weave data at {urls.project_weave_root_url(entity_name, project_name)}"
    )
    # Cosmetically, if we are in `read_only` mode, we are not logging data, so
    # we should not print the message about logging data.
    if not read_only:
        print(message)



import os

import pytest
from openai import AsyncOpenAI, OpenAI

import weave
from weave.integrations.integration_utilities import op_name_from_ref

model = "gpt-4o"


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_openai_quickstart(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    response = openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "How are you?"}],
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "I'm just a computer program, so I don't have feelings, but I'm here and ready to help you! How can I assist you today?"
    assert response.choices[0].message.content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["requests"] == 1
    assert usage["completion_tokens"] == 28
    assert usage["prompt_tokens"] == 11
    assert usage["total_tokens"] == 39

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [{"role": "user", "content": "How are you?"}]
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_async_quickstart(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "How are you?"}],
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "I'm just a computer program, so I don't have feelings, but I'm here and ready to help you! How can I assist you today?"
    assert response.choices[0].message.content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["requests"] == 1
    assert usage["completion_tokens"] == 28
    assert usage["prompt_tokens"] == 11
    assert usage["total_tokens"] == 39

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [{"role": "user", "content": "How are you?"}]
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_openai_stream_quickstart(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    response = openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "How are you?"}],
        temperature=0.0,
        max_tokens=64,
        top_p=1,
        stream=True,
    )

    all_content = ""
    for chunk in response:
        if chunk.choices[0].delta.content:
            all_content += chunk.choices[0].delta.content

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "I'm just a computer program, so I don't have feelings, but thanks for asking! How can I assist you today?"
    assert all_content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output["model"] == "gpt-4o-2024-05-13"
    assert output["object"] == "chat.completion"

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [{"role": "user", "content": "How are you?"}]
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1

    # usage information should be available even if `stream_options` is not set
    usage = call.summary["usage"][output["model"]]  # type: ignore
    assert usage["total_tokens"] == 35
    assert usage["completion_tokens"] == 24
    assert usage["prompt_tokens"] == 11

    # since we are setting `stream_options`, the chunk should not have usage information
    assert chunk.usage is None


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_async_stream_quickstart(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "How are you?"}],
        temperature=0.0,
        max_tokens=64,
        top_p=1,
        stream=True,
    )
    all_content = ""
    async for chunk in response:
        if chunk.choices[0].delta.content:
            all_content += chunk.choices[0].delta.content

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "I'm just a computer program, so I don't have feelings, but thanks for asking! How can I assist you today?"
    assert all_content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output["model"] == "gpt-4o-2024-05-13"
    assert output["object"] == "chat.completion"

    usage = call.output["usage"]  # type: ignore
    assert usage["total_tokens"] == 35
    assert usage["completion_tokens"] == 24
    assert usage["prompt_tokens"] == 11
    assert chunk.usage is None

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [{"role": "user", "content": "How are you?"}]
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_openai_stream_usage_quickstart(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    response = openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "How are you?"}],
        temperature=0.0,
        max_tokens=64,
        top_p=1,
        stream=True,
        stream_options={
            "include_usage": True
        },  # User needs to pass this argument to get usage
    )

    for chunk in response:
        pass
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    usage = call.summary["usage"][output["model"]]  # type: ignore
    assert usage["total_tokens"] == 35
    assert usage["completion_tokens"] == 24
    assert usage["prompt_tokens"] == 11


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_openai_function_call(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    function_list = [
        {
            "name": "cricket_player_names",  # Function Name
            "description": "store the name of players",  # Meta information of function
            "parameters": {  # parameters
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "The name of the player",
                    },
                    "team:": {
                        "type": "string",
                        "description": "The team of the player",
                    },
                    "highest_score": {
                        "type": "number",
                        "description": "The highest score of the player",
                    },
                },
                "required": ["name", "team", "highest_score"],
            },
        }
    ]

    response = openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        functions=function_list,
        function_call={
            "name": "cricket_player_names"
        },  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    print(response)
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":248}'
    assert response.choices[0].message.function_call.arguments == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["total_tokens"] == 117
    assert usage["completion_tokens"] == 18
    assert usage["prompt_tokens"] == 99

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["function_call"]["name"] == "cricket_player_names"
    assert inputs["functions"] == function_list
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_function_call_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    function_list = [
        {
            "name": "cricket_player_names",  # Function Name
            "description": "store the name of players",  # Meta information of function
            "parameters": {  # parameters
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "The name of the player",
                    },
                    "team:": {
                        "type": "string",
                        "description": "The team of the player",
                    },
                    "highest_score": {
                        "type": "number",
                        "description": "The highest score of the player",
                    },
                },
                "required": ["name", "team", "highest_score"],
            },
        }
    ]

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        functions=function_list,
        function_call={
            "name": "cricket_player_names"
        },  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":200}'
    assert response.choices[0].message.function_call.arguments == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["total_tokens"] == 117
    assert usage["completion_tokens"] == 18
    assert usage["prompt_tokens"] == 99

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["function_call"]["name"] == "cricket_player_names"
    assert inputs["functions"] == function_list
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_function_call_async_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    function_list = [
        {
            "name": "cricket_player_names",  # Function Name
            "description": "store the name of players",  # Meta information of function
            "parameters": {  # parameters
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "The name of the player",
                    },
                    "team:": {
                        "type": "string",
                        "description": "The team of the player",
                    },
                    "highest_score": {
                        "type": "number",
                        "description": "The highest score of the player",
                    },
                },
                "required": ["name", "team", "highest_score"],
            },
        }
    ]

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        functions=function_list,
        function_call={
            "name": "cricket_player_names"
        },  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
        stream=True,
    )
    # TODO: figure out if this is the write pattern
    async for chunk in response:
        pass

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":200}'
    choice = call.output["choices"][0]  # type: ignore
    assert choice["message"]["function_call"]["arguments"] == exp
    assert choice["message"]["function_call"]["name"] == "cricket_player_names"
    assert choice["finish_reason"] == "stop"
    assert choice["message"]["role"] == "assistant"

    assert call.output["id"] == "chatcmpl-9cDoAspZs24R7hQPixbNse9Lg5dgR"  # type: ignore
    assert call.output["model"] == "gpt-4o-2024-05-13"  # type: ignore
    assert call.output["object"] == "chat.completion"  # type: ignore

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["function_call"]["name"] == "cricket_player_names"
    assert inputs["functions"] == function_list
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_openai_tool_call(client: weave.trace.weave_client.WeaveClient) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    function_list = [
        {
            "type": "function",
            "function": {
                "name": "cricket_player_names",  # Function Name
                "description": "store the name of players",  # Meta information of function
                "parameters": {  # parameters
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "The name of the player",
                        },
                        "team:": {
                            "type": "string",
                            "description": "The team of the player",
                        },
                        "highest_score": {
                            "type": "number",
                            "description": "The highest score of the player",
                        },
                    },
                    "required": ["name", "team", "highest_score"],
                },
            },
        }
    ]

    response = openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        tools=function_list,
        tool_choice="required",  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    print(response)

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":248}'
    assert response.choices[0].message.tool_calls[0].function.arguments == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["total_tokens"] == 117
    assert usage["completion_tokens"] == 27
    assert usage["prompt_tokens"] == 90

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["tools"] == function_list
    assert inputs["tool_choice"] == "required"
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_tool_call_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    function_list = [
        {
            "type": "function",
            "function": {
                "name": "cricket_player_names",  # Function Name
                "description": "store the name of players",  # Meta information of function
                "parameters": {  # parameters
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "The name of the player",
                        },
                        "team:": {
                            "type": "string",
                            "description": "The team of the player",
                        },
                        "highest_score": {
                            "type": "number",
                            "description": "The highest score of the player",
                        },
                    },
                    "required": ["name", "team", "highest_score"],
                },
            },
        }
    ]

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        tools=function_list,
        tool_choice="required",  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":248}'
    assert response.choices[0].message.tool_calls[0].function.arguments == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output.model == "gpt-4o-2024-05-13"
    assert output.object == "chat.completion"

    usage = call.summary["usage"][output.model]  # type: ignore
    assert usage["total_tokens"] == 117
    assert usage["completion_tokens"] == 27
    assert usage["prompt_tokens"] == 90

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["tools"] == function_list
    assert inputs["tool_choice"] == "required"
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_tool_call_async_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    function_list = [
        {
            "type": "function",
            "function": {
                "name": "cricket_player_names",  # Function Name
                "description": "store the name of players",  # Meta information of function
                "parameters": {  # parameters
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "The name of the player",
                        },
                        "team:": {
                            "type": "string",
                            "description": "The team of the player",
                        },
                        "highest_score": {
                            "type": "number",
                            "description": "The highest score of the player",
                        },
                    },
                    "required": ["name", "team", "highest_score"],
                },
            },
        }
    ]

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": "Can you name a cricket player along with team name and highest score?",
            }
        ],
        tools=function_list,
        tool_choice="required",  # this forces the LLM to do function call.
        temperature=0.0,
        max_tokens=64,
        top_p=1,
        stream=True,
        stream_options={
            "include_usage": True
        },  # User needs to pass this argument to get usage
    )
    # TODO: figure out if this is the write pattern
    async for chunk in response:
        pass

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    exp = '{"name":"Sachin Tendulkar","team":"India","highest_score":248}'
    choice = call.output["choices"][0]  # type: ignore
    assert choice["message"]["tool_calls"][0]["function"]["arguments"] == exp
    assert (
        choice["message"]["tool_calls"][0]["function"]["name"] == "cricket_player_names"
    )
    assert choice["message"]["tool_calls"][0]["id"] == "call_f0b9z5IxIGxTMHYa55QnIXdG"
    assert choice["message"]["tool_calls"][0]["type"] == "function"
    assert choice["finish_reason"] == "tool_calls"
    assert choice["message"]["role"] == "assistant"

    assert call.output["id"] == "chatcmpl-9cEL2tGY6V4efCUFlecqgb6HZOhpV"  # type: ignore
    assert call.output["model"] == "gpt-4o-2024-05-13"  # type: ignore
    assert call.output["object"] == "chat.completion"  # type: ignore

    usage = call.output["usage"]  # type: ignore
    assert usage["total_tokens"] == 117
    assert usage["completion_tokens"] == 27
    assert usage["prompt_tokens"] == 90

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert (
        inputs["messages"][0]["content"]
        == "Can you name a cricket player along with team name and highest score?"
    )
    assert inputs["tools"] == function_list
    assert inputs["tool_choice"] == "required"
    assert inputs["max_tokens"] == 64
    assert inputs["temperature"] == 0.0
    assert inputs["top_p"] == 1


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_openai_as_context_manager(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = OpenAI(api_key=api_key)

    with openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "Hello, I am context manager!"}],
        stream=True,
    ) as response:
        all_content = ""
        for chunk in response:
            if chunk.choices[0].delta.content:
                all_content += chunk.choices[0].delta.content

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "Hello! As a context manager, you play a crucial role in managing resources efficiently and ensuring proper setup and teardown in Python. You likely implement methods like `__enter__` and `__exit__` to manage contexts properly. How can I assist you today?"
    assert all_content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output["model"] == "gpt-4o-2024-05-13"
    assert output["object"] == "chat.completion"

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [
        {"role": "user", "content": "Hello, I am context manager!"}
    ]

    # usage information should be available even if `stream_options` is not set
    usage = call.summary["usage"][output["model"]]  # type: ignore
    assert usage["total_tokens"] == 67
    assert usage["completion_tokens"] == 53
    assert usage["prompt_tokens"] == 14

    # since we are setting `stream_options`, the chunk should not have usage information
    assert chunk.usage is None


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_openai_as_context_manager_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")

    openai_client = AsyncOpenAI(api_key=api_key)

    response = await openai_client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": "Hello, I am async context manager!"}],
        stream=True,
    )

    async with response:
        all_content = ""
        async for chunk in response:
            if chunk.choices[0].delta.content:
                all_content += chunk.choices[0].delta.content

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    exp = "Hello! It sounds like you're referring to the concept of an asynchronous context manager in programming, typically found in languages like Python. Asynchronous context managers are useful for managing resources that need to be setup and cleaned up, often in an asynchronous manner. Do you have any specific questions or scenarios you need help with regarding async context managers?"
    assert all_content == exp

    assert op_name_from_ref(call.op_name) == "openai.chat.completions.create"
    assert call.started_at is not None
    assert call.started_at < call.ended_at  # type: ignore

    output = call.output
    assert output["model"] == "gpt-4o-2024-05-13"
    assert output["object"] == "chat.completion"

    inputs = call.inputs
    assert inputs["model"] == "gpt-4o"
    assert inputs["messages"] == [
        {"role": "user", "content": "Hello, I am async context manager!"}
    ]

    # usage information should be available even if `stream_options` is not set
    usage = call.summary["usage"][output["model"]]  # type: ignore
    assert usage["total_tokens"] == 81
    assert usage["completion_tokens"] == 66
    assert usage["prompt_tokens"] == 15

    # since we are setting `stream_options`, the chunk should not have usage information
    assert chunk.usage is None



# Safe class for managing W&B Artifact downloads. Can be used
# as the core of a fast artifact downloader, or a safe multiplexed
# file manager in the Weave server.

import datetime
import json
import typing
import urllib
from urllib import parse

from aiohttp import BasicAuth
from requests.auth import HTTPBasicAuth
from wandb.sdk.lib import hashutil

from weave.legacy.weave import environment as weave_env
from weave.legacy.weave import filesystem, artifact_wandb, cache, errors, wandb_api, engine_trace, weave_http

tracer = engine_trace.tracer()  # type: ignore


def _file_path(
    uri: typing.Union[
        artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
    ],
    md5_hex: str,
) -> str:
    if uri.path is None:
        raise errors.WeaveInternalError("Artifact URI has no path in call to file_path")
    path_parts = uri.path.split(".", 1)
    if len(path_parts) == 2:
        extension = "." + path_parts[1]
    else:
        extension = ""
    if isinstance(uri, artifact_wandb.WeaveWBArtifactByIDURI):
        return f"wandb_file_manager/{uri.path_root}/{uri.artifact_id}/{uri.name}/{md5_hex}{extension}"
    return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/{md5_hex}{extension}"


def _local_path_and_download_url(
    art_uri: typing.Union[
        artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
    ],
    manifest: artifact_wandb.WandbArtifactManifest,
    base_url: typing.Optional[str] = None,
) -> typing.Optional[typing.Tuple[str, str]]:
    path = art_uri.path
    if path is None:
        raise errors.WeaveInternalError(
            "Artifact URI has no path in call to _local_path_and_download_url"
        )
    file_name = path.split("/")[-1]
    manifest_entry = manifest.get_entry_by_path(path)
    if manifest_entry is None:
        return None
    md5_hex = hashutil.b64_to_hex_id(hashutil.B64MD5(manifest_entry["digest"]))
    base_url = base_url or weave_env.wandb_base_url()
    file_path = _file_path(art_uri, md5_hex)
    entity_name = urllib.parse.quote(art_uri.entity_name or "_")
    # For artifactsV1 (legacy artifacts), the artifact's files cannot be fetched without the entity name, but
    # we substitute a '_' here anyway since artifact_uri.entity_name can be None accessed via an organization's
    # registry collection.
    if manifest.storage_layout == artifact_wandb.WandbArtifactManifest.StorageLayout.V1:
        return file_path, "{}/artifacts/{}/{}/{}".format(
            base_url, entity_name or "_", md5_hex, urllib.parse.quote(file_name)
        )
    else:
        # TODO: storage_region
        storage_region = "default"
        # For artifactsV2 (which is all artifacts now), the file download handler ignores the entity
        # parameter while parsing the url, and fetches the files directly via the artifact id
        # Refer to: https://github.com/wandb/core/blob/7cfee1cd07ddc49fe7ba70ce3d213d2a11bd4456/services/gorilla/api/handler/artifacts.go#L179
        return file_path, "{}/artifactsV2/{}/{}/{}/{}/{}".format(
            base_url,
            storage_region,
            entity_name or "_",
            urllib.parse.quote(manifest_entry.get("birthArtifactID", "")),  # type: ignore
            md5_hex,
            urllib.parse.quote(file_name),
        )


class WandbFileManagerAsync:
    def __init__(
        self,
        filesystem: filesystem.FilesystemAsync,
        http: weave_http.HttpAsync,
        wandb_api: wandb_api.WandbApiAsync,
    ) -> None:
        self.fs = filesystem
        self.http = http
        self.wandb_api = wandb_api
        self._manifests: cache.LruTimeWindowCache[
            str, typing.Optional[artifact_wandb.WandbArtifactManifest]
        ] = cache.LruTimeWindowCache(datetime.timedelta(minutes=5))

    def manifest_path(
        self,
        uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> str:
        assert uri.version is not None
        if isinstance(uri, artifact_wandb.WeaveWBArtifactURI):
            return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/manifest-{uri.version}.json"
        return f"wandb_file_manager/{uri.path_root}/{uri.artifact_id}/{uri.name}/manifest-{uri.version}.json"

    async def _manifest(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
        manifest_path: str,
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        if art_uri.version is None:
            raise errors.WeaveInternalError(
                'Artifact URI has no version: "%s"' % art_uri
            )
        # Check if on disk
        try:
            async with self.fs.open_read(manifest_path) as f:
                return artifact_wandb.WandbArtifactManifest(json.loads(await f.read()))
        except FileNotFoundError:
            pass
        # Download
        manifest_url = None
        if isinstance(art_uri, artifact_wandb.WeaveWBArtifactByIDURI):
            artifact_id = art_uri.artifact_id
            manifest_url = await self.wandb_api.artifact_manifest_url_from_id(
                art_id=artifact_id,
            )
        else:
            manifest_url = await self.wandb_api.artifact_manifest_url(
                art_uri.entity_name,
                art_uri.project_name,
                art_uri.name + ":" + art_uri.version,
            )
        if manifest_url is None:
            return None
        await self.http.download_file(
            manifest_url,
            manifest_path,
        )
        async with self.fs.open_read(manifest_path) as f:
            return artifact_wandb.WandbArtifactManifest(json.loads(await f.read()))

    async def manifest(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        with tracer.trace("wandb_file_manager.manifest") as span:
            assert art_uri.version is not None
            manifest_path = self.manifest_path(art_uri)
            manifest = self._manifests.get(manifest_path)
            if not isinstance(manifest, cache.LruTimeWindowCache.NotFound):
                return manifest
            manifest = await self._manifest(art_uri, manifest_path)
            self._manifests.set(manifest_path, manifest)
            return manifest

    async def local_path_and_download_url(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
        base_url: typing.Optional[str] = None,
    ) -> typing.Optional[typing.Tuple[str, str]]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to local_path_and_download_url"
            )
        manifest = await self.manifest(art_uri)
        if manifest is None:
            return None
        return _local_path_and_download_url(art_uri, manifest, base_url)

    async def ensure_file(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file") as span:
            span.set_tag("uri", str(art_uri))
            res = await self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            file_path, download_url = res
            if await self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = BasicAuth("api", wandb_api_context.api_key)
            await self.http.download_file(
                download_url,
                file_path,
                headers=headers,
                cookies=cookies,
                auth=auth,
            )
            return file_path

    async def ensure_file_downloaded(
        self,
        download_url: str,
    ) -> typing.Optional[str]:
        """Ensures a history parquet file from an http/https URI."""
        schema, netloc, path, _, _, _ = parse.urlparse(download_url)
        path = "/".join([netloc, path.lstrip("/")])
        if schema not in ("http", "https"):
            raise errors.WeaveInternalError(
                "ensure_file_downloaded only supports http/https URIs"
            )
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file_downloaded") as span:
            span.set_tag("download_url", str(download_url))
            file_path = f"wandb_file_manager/{path}"
            if await self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = BasicAuth("api", wandb_api_context.api_key)
            await self.http.download_file(
                download_url,
                file_path,
                headers=headers,
                cookies=cookies,
                auth=auth,
            )
            return file_path

    async def direct_url(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.direct_url") as span:
            span.set_tag("uri", str(art_uri))
            res = await self.local_path_and_download_url(
                art_uri, base_url=weave_env.wandb_frontend_base_url()
            )
            if res is None:
                return None
            return res[1]


class WandbFileManager:
    def __init__(
        self,
        filesystem: filesystem.Filesystem,
        http: weave_http.Http,
        wandb_api: wandb_api.WandbApi,
    ) -> None:
        self.fs = filesystem
        self.http = http
        self.wandb_api = wandb_api
        self._manifests: cache.LruTimeWindowCache[
            str, typing.Optional[artifact_wandb.WandbArtifactManifest]
        ] = cache.LruTimeWindowCache(datetime.timedelta(minutes=5))

    def manifest_path(
        self,
        uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> str:
        assert uri.version is not None
        if isinstance(uri, artifact_wandb.WeaveWBArtifactByIDURI):
            return f"wandb_file_manager/{uri.path_root}/{uri.artifact_id}/{uri.name}/manifest-{uri.version}.json"
        return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/manifest-{uri.version}.json"

    def _manifest(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
        manifest_path: str,
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        if art_uri.version is None:
            raise errors.WeaveInternalError(
                'Artifact URI has no version: "%s"' % art_uri
            )
        # Check if on disk
        try:
            with self.fs.open_read(manifest_path) as f:
                return artifact_wandb.WandbArtifactManifest(json.loads(f.read()))
        except FileNotFoundError:
            pass
        # Download
        manifest_url = None
        if isinstance(art_uri, artifact_wandb.WeaveWBArtifactByIDURI):
            artifact_id = art_uri.artifact_id
            manifest_url = self.wandb_api.artifact_manifest_url_from_id(
                art_id=artifact_id,
            )
        else:
            manifest_url = self.wandb_api.artifact_manifest_url(
                art_uri.entity_name,
                art_uri.project_name,
                art_uri.name + ":" + art_uri.version,
            )
        if manifest_url is None:
            return None
        self.http.download_file(
            manifest_url,
            manifest_path,
        )
        with self.fs.open_read(manifest_path) as f:
            return artifact_wandb.WandbArtifactManifest(json.loads(f.read()))

    def manifest(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        with tracer.trace("wandb_file_manager.manifest") as span:
            assert art_uri.version is not None
            manifest_path = self.manifest_path(art_uri)
            manifest = self._manifests.get(manifest_path)
            if not isinstance(manifest, cache.LruTimeWindowCache.NotFound):
                return manifest
            manifest = self._manifest(art_uri, manifest_path)
            self._manifests.set(manifest_path, manifest)
            return manifest

    def local_path_and_download_url(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
        base_url: typing.Optional[str] = None,
    ) -> typing.Optional[typing.Tuple[str, str]]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to local_path_and_download_url"
            )
        manifest = self.manifest(art_uri)
        if manifest is None:
            return None
        return _local_path_and_download_url(art_uri, manifest, base_url)

    def ensure_file_downloaded(
        self,
        download_url: str,
    ) -> typing.Optional[str]:
        """Ensures a history parquet file from an http/https URI."""
        schema, netloc, path, _, _, _ = parse.urlparse(download_url)
        path = "/".join([netloc, path.lstrip("/")])
        if schema not in ("http", "https"):
            raise errors.WeaveInternalError(
                "ensure_file_downloaded only supports http/https URIs"
            )
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file_downloaded") as span:
            span.set_tag("download_url", str(download_url))
            file_path = f"wandb_file_manager/{path}"
            if self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = HTTPBasicAuth("api", wandb_api_context.api_key)
            self.http.download_file(
                download_url, file_path, headers=headers, cookies=cookies, auth=auth
            )
            return file_path

    def ensure_file(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file") as span:
            span.set_tag("uri", str(art_uri))
            res = self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            file_path, download_url = res
            if self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = HTTPBasicAuth("api", wandb_api_context.api_key)
            self.http.download_file(
                download_url, file_path, headers=headers, cookies=cookies, auth=auth
            )
            return file_path

    def direct_url(
        self,
        art_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.direct_url") as span:
            span.set_tag("uri", str(art_uri))
            res = self.local_path_and_download_url(
                art_uri, base_url=weave_env.wandb_frontend_base_url()
            )
            if res is None:
                return None
            return res[1]



import typing

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op


@op(
    name="type-name",
)
def type_name(self: types.Type) -> str:
    return self.name


def _cast_output_type(input_type):
    if not isinstance(input_type["to_type"], types.Const):
        raise errors.WeaveTypeError(f"Expected Const, got {input_type}")
    return input_type["to_type"].val


# This is not yet supported in js. The right argument needs to be Const,
# but we don't have a way to construct const objects in js yet.
@op(
    input_type={"to_type": types.Const(types.Type(), None)},
    output_type=_cast_output_type,
)
def cast(obj: typing.Any, to_type):
    obj_type = types.TypeRegistry.type_of(obj)
    if not to_type.assign_type(obj_type):
        raise errors.WeaveTypeError(f"Cannot cast {obj_type} to {to_type}")
    return obj



import os
import re

import pytest
import requests

from weave import weave_server
from weave.legacy.weave import api, context, logs, ops, server


def test_logfile_created(fresh_server_logfile):
    server.capture_weave_server_logs()
    with context.local_http_client():
        # run this to kick off a server
        assert api.use(ops.Number.__add__(3, 9)) == 12

        # check that the log file was created
        log_filename = logs.default_log_filename()
        assert log_filename
        assert os.path.exists(log_filename)

        with open(log_filename, "r") as f:
            content = f.read()

        # check that it has a record of executing one call
        assert "execute" in content and "200" in content


def test_logfile_captures_error(fresh_server_logfile):
    # run this to kick off a server
    server.capture_weave_server_logs()

    with context.local_http_client():
        with pytest.raises(requests.exceptions.HTTPError):
            api.use(ops.Number.__add__(4, "a"))

        # check that the log file was created
        log_filename = logs.default_log_filename()
        assert log_filename
        assert os.path.exists(log_filename)

        with open(log_filename, "r") as f:
            content = f.read()

        # check that it has a record of executing one call
        assert "execute" in content and "500" in content


def test_log_2_app_instances_different_threads(fresh_server_logfile):
    # kick off two http servers
    server.capture_weave_server_logs()
    with context.local_http_client():
        with context.local_http_client():
            with pytest.raises(requests.exceptions.HTTPError):
                # this one is run by server 2
                api.use(ops.Number.__add__(3, "a"))

        # this one is run by server 1
        assert api.use(ops.Number.__add__(3, 9)) == 12

    # check that the log file was created
    log_filename = logs.default_log_filename()
    assert log_filename
    assert os.path.exists(log_filename)

    with open(log_filename, "r") as f:
        content = f.read()

    # check that it has a record of executing one call
    assert "execute" in content
    assert "200" in content

    # and we have an error
    assert "500" in content

    # check that there are two threads logged
    threads = set(re.findall(r"Weave Port: (.+?)\)", content))
    assert len(threads) == 2


def test_capture_server_logs_captures_server_logs(fresh_server_logfile, capsys):
    server.capture_weave_server_logs()

    with context.local_http_client():
        # run this to kick off a server
        assert api.use(ops.Number.__add__(3, 9)) == 12

        # check that the log file was created
        log_filename = logs.default_log_filename()
        assert log_filename
        assert os.path.exists(log_filename)

        with open(log_filename, "r") as f:
            content = f.read()

        # check that it has a record of executing one call
        assert "execute" in content and "200" in content

        # check that the same stuff appears in the captured stream logs
        capresult = capsys.readouterr()
        stderr = capresult.err
        assert "execute" in stderr and "200" in stderr



import asyncio
import threading

import aioprocessing
import pytest

from weave.legacy.weave.async_queue import ProcessQueue, Queue, ThreadQueue


async def process_producer(queue: Queue) -> None:
    tasks = set()
    loop = asyncio.get_running_loop()
    for i in range(3):
        print(f"Producer: {i}", flush=True)
        task = loop.create_task(queue.async_put(i))
        tasks.add(task)
    await asyncio.wait(tasks)


def process_consumer(queue: Queue) -> None:
    async def _consume():
        for _ in range(3):
            if isinstance(queue, ThreadQueue):
                item = queue.get()
            else:
                item = await queue.async_get()
            print(f"Consumer: {item}", flush=True)
            queue.task_done()

    asyncio.run(_consume())


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_async_process_queue_shared() -> None:
    queue: Queue = ProcessQueue()

    consumer_process = aioprocessing.AioProcess(target=process_consumer, args=(queue,))
    consumer_process.start()

    await process_producer(queue)

    queue.join()
    consumer_process.join()


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_async_thread_queue_shared() -> None:
    queue: Queue = ThreadQueue()
    consumer_thread = threading.Thread(target=process_consumer, args=(queue,))
    consumer_thread.start()

    await process_producer(queue)
    queue.join()
    consumer_thread.join()



import typing

import graphql

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import gql_op_plugin, gql_to_weave, graph, op_args, stitch, registry_mem
from weave.legacy.weave.input_provider import InputAndStitchProvider

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def


def fragment_to_query(fragment: str) -> str:
    return f"query WeavePythonCG {{ {fragment} }}"


# This is the primary exposed function of this module and is called in `compile.py`. It's primary role
# is to roll up all the GQLOps such that a single query can be composed - then replace the root op with a
# special op that calls such query and constructs the correct type. It makes heavy use of the
# `stitch` module to do this. Moreover, all GQL is zipped and deduped so that the minimum request
# is made to the server. See the helper functions below for more details.
def apply_domain_op_gql_translation(
    leaf_nodes: list[graph.Node], on_error: graph.OnErrorFnType = None
) -> list[graph.Node]:
    # Only apply this transformation at least one of the leaf nodes is a root node
    if not graph.filter_nodes_full(leaf_nodes, _is_root_node):
        return leaf_nodes

    p = stitch.stitch(leaf_nodes, on_error=on_error)

    query_str_const_node = graph.ConstNode(types.String(), "")
    alias_list_const_node = graph.ConstNode(types.List(types.String()), [])
    output_type_node = graph.ConstNode(types.TypeType(), types.Any())
    query_root_node = graph.OutputNode(
        types.Dict(types.String(), types.TypedDict({})),
        "gqlroot-wbgqlquery",
        {
            "query_str": query_str_const_node,
            "alias_list": alias_list_const_node,
            "output_type": output_type_node,
        },
    )
    fragments = []
    aliases = []

    def _replace_with_merged_gql(node: graph.Node) -> graph.Node:
        if not _is_root_node(node):
            return node
        node = typing.cast(graph.OutputNode, node)
        inner_fragment = _get_fragment(node, p)
        fragments.append(inner_fragment)
        alias = gql_to_weave.get_outermost_alias(inner_fragment)
        aliases.append(alias)
        custom_resolver = _custom_root_resolver(node)

        if custom_resolver is not None:
            return custom_resolver(query_root_node, **node.from_op.inputs)
        else:
            output_type = get_plugin_output_type(node)

            return graph.OutputNode(
                output_type,
                "gqlroot-querytoobj",
                {
                    "result_dict": query_root_node,
                    "result_key": graph.ConstNode(types.String(), alias),
                    "output_type": graph.ConstNode(types.TypeType(), output_type),
                    "gql_query_fragment": graph.ConstNode(
                        types.String(), inner_fragment
                    ),
                },
            )

    res = graph.map_nodes_full(leaf_nodes, _replace_with_merged_gql, on_error)

    combined_query_fragment = "\n".join(fragments)
    query_str = fragment_to_query(combined_query_fragment)
    if combined_query_fragment.strip() != "":
        query_str = normalize_gql_query_string(query_str)
    query_str_const_node.val = query_str
    alias_list_const_node.val = aliases

    return res


### Everything below are helpers for the above function ###


def get_plugin_output_type(node: graph.OutputNode) -> types.Type:
    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    return op_def.concrete_output_type


def _get_fragment(node: graph.OutputNode, stitchedGraph: stitch.StitchedGraph) -> str:
    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    # TODO: make this a helper (it is used in stich.py as well)
    if op_def.derived_from and op_def.derived_from.derived_ops["mapped"] == op_def:
        op_def = op_def.derived_from

    # These are all passthrough ops - should this be in stitch?
    is_passthrough = (
        op_def.name.endswith("offset")
        or op_def.name.endswith("limit")
        or op_def.name.endswith("index")
        or op_def.name.endswith("__getitem__")
        or op_def.name.endswith("concat")
        or op_def.name.endswith("contains")
        or op_def.name.endswith("list")
        or op_def.name.endswith("concat")
        or op_def.name.endswith("flatten")
        or op_def.name.endswith("unnest")
        or op_def.name.endswith("dropna")
        or op_def.name == "list-createIndexCheckpointTag"
    )

    wb_domain_gql = gql_op_plugin.get_gql_plugin(op_def)
    if wb_domain_gql is None and not is_passthrough:
        return ""

    forward_obj = stitchedGraph.get_result(node)
    calls = forward_obj.calls
    child_fragment = "\n".join(
        [
            _get_fragment(call.node, stitchedGraph)
            for call in calls
            if isinstance(call.node, graph.OutputNode)
        ]
    )

    if is_passthrough:
        return child_fragment
    wb_domain_gql = typing.cast(gql_op_plugin.GqlOpPlugin, wb_domain_gql)

    const_node_input_vals = {
        key: value.val
        for key, value in node.from_op.inputs.items()
        if isinstance(value, graph.ConstNode)
    }
    ip = InputAndStitchProvider(const_node_input_vals, forward_obj)

    fragment = wb_domain_gql.query_fn(
        ip,
        child_fragment,
    )

    return fragment


def _get_configsummaryhistory_keys_or_specs(
    field: graphql.language.ast.FieldNode,
) -> typing.Optional[typing.List[str]]:
    if len(field.arguments) > 1:
        raise errors.WeaveInternalError(
            "Custom merge for config, summaryMetrics, and sampledHistory only supports one argument"
        )
    if not field.arguments:
        return None
    field_arg0 = field.arguments[0]
    if field_arg0.name.value not in ["keys", "specs", "liveKeys"]:
        raise errors.WeaveInternalError(
            "First argument for custom merge must be 'keys' or 'specs'"
        )
    if not isinstance(field_arg0.value, graphql.language.ast.ListValueNode):
        raise errors.WeaveInternalError(
            "First argument for custom merge must be a list"
        )
    keys: list[str] = []
    for key in field_arg0.value.values:
        if not isinstance(key, graphql.language.ast.StringValueNode):
            raise errors.WeaveInternalError(
                "First argument for custom merge must be a list of strings"
            )
        keys.append(key.value)
    return keys


def _field_selections_hardcoded_merge(
    merge_from: graphql.language.ast.FieldNode,
    merge_to: graphql.language.ast.FieldNode,
) -> bool:
    # Custom field merging for config, summaryMetrics, and history keys.
    # Must be kept in sync with run_ops config* and summary_metrics* ops.
    if merge_from.name.value != merge_to.name.value:
        return False
    if merge_from.name.value not in [
        "config",
        "summaryMetrics",
        "sampledHistory",
        "parquetHistory",
    ]:
        return False
    if (
        merge_from.alias is None
        or merge_to.alias is None
        or merge_from.alias.value != merge_to.alias.value
    ):
        return False
    if (
        merge_from.alias.value != "configSubset"
        and merge_from.alias.value != "summaryMetricsSubset"
        and merge_from.alias.value != "sampledHistorySubset"
        and merge_from.alias.value != "sampledParquetHistory"
    ):
        return False
    merge_from_keys = _get_configsummaryhistory_keys_or_specs(merge_from)
    merge_to_keys = _get_configsummaryhistory_keys_or_specs(merge_to)
    if merge_to_keys is None:
        # merge_to already selects all
        pass
    elif merge_from_keys is None:
        # select all by removing arguments
        merge_to.arguments = ()
    else:
        for k in merge_from_keys:
            if k not in merge_to_keys:
                merge_to_keys.append(k)
        merge_to.arguments[0].value.values = tuple(
            graphql.language.ast.StringValueNode(value=k) for k in merge_to_keys
        )
    return True


def _field_selections_are_mergeable(
    selection1: graphql.language.ast.FieldNode,
    selection2: graphql.language.ast.FieldNode,
) -> bool:
    if selection1.name.value != selection2.name.value:
        return False
    if selection1.alias is None and selection2.alias is not None:
        return False
    if selection1.alias is not None and selection2.alias is None:
        return False
    if (
        selection1.alias is not None
        and selection2.alias is not None
        and selection1.alias.value != selection2.alias.value
    ):
        return False
    if len(selection1.arguments) != len(selection2.arguments):
        return False
    for arg1, arg2 in zip(selection1.arguments, selection2.arguments):
        if arg1.name.value != arg2.name.value:
            return False
        if arg1.value.to_dict() != arg2.value.to_dict():
            return False
    return True


def _fragment_selections_are_mergeable(
    selection1: graphql.language.ast.InlineFragmentNode,
    selection2: graphql.language.ast.InlineFragmentNode,
) -> bool:
    return selection1.type_condition.name.value == selection2.type_condition.name.value


def _zip_selection_set(
    selection_set: graphql.language.ast.SelectionSetNode,
) -> graphql.language.ast.SelectionSetNode:
    selections = selection_set.selections
    # Two selections can be merged if their alias, name, and arguments are the same
    # unfortunately, this seems to be O(n^2) in the number of selections
    new_selections: list[
        typing.Union[
            graphql.language.ast.FieldNode, graphql.language.ast.InlineFragmentNode
        ]
    ] = []
    for selection in selections:
        if isinstance(selection, graphql.language.ast.FieldNode) or isinstance(
            selection, graphql.language.ast.InlineFragmentNode
        ):
            for new_selection in new_selections:
                did_custom_merge = (
                    isinstance(selection, graphql.language.ast.FieldNode)
                    and isinstance(new_selection, graphql.language.ast.FieldNode)
                    and _field_selections_hardcoded_merge(selection, new_selection)
                )
                if did_custom_merge:
                    break

                should_merge = (
                    isinstance(selection, graphql.language.ast.FieldNode)
                    and isinstance(new_selection, graphql.language.ast.FieldNode)
                    and _field_selections_are_mergeable(selection, new_selection)
                )
                should_merge = should_merge or (
                    isinstance(selection, graphql.language.ast.InlineFragmentNode)
                    and isinstance(
                        new_selection, graphql.language.ast.InlineFragmentNode
                    )
                    and _fragment_selections_are_mergeable(selection, new_selection)
                )
                if should_merge:
                    if selection.selection_set:
                        if new_selection.selection_set is None:
                            new_selection.selection_set = (
                                graphql.language.ast.SelectionSetNode(selections=())
                            )
                        new_selection.selection_set.selections = (
                            *new_selection.selection_set.selections,
                            *selection.selection_set.selections,
                        )
                    break
            else:
                new_selections.append(selection)
        else:
            raise ValueError(
                f"Found unsupported selection type {type(selection)}, please add implementation in compile_domain.py"
            )
    for selection in new_selections:
        if selection.selection_set:
            selection.selection_set = _zip_selection_set(selection.selection_set)
    selection_set.selections = tuple(new_selections)
    return selection_set


def _zip_gql_doc(
    gql_doc: graphql.language.ast.DocumentNode,
) -> graphql.language.ast.DocumentNode:
    if len(gql_doc.definitions) > 1:
        raise ValueError("Only one query definition is supported")
    query_def = gql_doc.definitions[0]
    if not isinstance(query_def, graphql.language.ast.OperationDefinitionNode):
        raise ValueError("Only operation definitions are supported")

    query_def.selection_set = _zip_selection_set(query_def.selection_set)
    gql_doc.definitions = (query_def,)
    return gql_doc


def normalize_gql_query_string(query_str: str) -> str:
    gql_doc = graphql.language.parse(query_str)
    gql_doc = _zip_gql_doc(gql_doc)
    return graphql.utilities.strip_ignored_characters(
        graphql.language.print_ast(gql_doc)
    )


def _is_root_node(node: graph.Node) -> bool:
    if not isinstance(node, graph.OutputNode):
        return False

    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    wb_domain_gql = gql_op_plugin.get_gql_plugin(op_def)
    return wb_domain_gql is not None and wb_domain_gql.is_root


def _custom_root_resolver(node: graph.Node) -> typing.Optional["op_def.OpDef"]:
    if not isinstance(node, graph.OutputNode):
        return None

    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    wb_domain_gql = gql_op_plugin.get_gql_plugin(op_def)
    if wb_domain_gql is not None:
        return wb_domain_gql.root_resolver
    return None


def required_const_input_names(node: graph.Node) -> typing.Optional[list[str]]:
    if not isinstance(node, graph.OutputNode):
        return None

    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    wb_domain_gql = gql_op_plugin.get_gql_plugin(op_def)
    if wb_domain_gql is None:
        return None
    if not isinstance(op_def.input_type, op_args.OpNamedArgs):
        raise errors.WeaveInternalError(
            'Only named args are supported for "gql" domain'
        )
    input_names = list(op_def.input_type.arg_types.keys())
    if wb_domain_gql.is_root:
        return input_names
    return input_names[1:]



# This is deprecated in favor of the new wandb_api.py module.
# TODO: remove uses of this and delete.

import logging
import typing

import graphql
from graphql import GraphQLSchema
from requests import exceptions
from wandb.apis import public
from wandb.apis.public.api import gql
from wandb.errors import CommError as WandbCommError
from wandb.sdk.internal.internal_api import _thread_local_api_settings

from weave.legacy.weave import errors


def wandb_public_api() -> public.Api:
    return public.Api(timeout=30)


def assert_wandb_authenticated() -> None:
    authenticated = (
        wandb_public_api().api_key is not None
        or _thread_local_api_settings.cookies is not None
    )
    if not authenticated:
        raise errors.WeaveWandbAuthenticationException(
            "Unable to log data to W&B. Please authenticate by setting WANDB_API_KEY or running `wandb init`."
        )


def query_with_retry(
    query_str: str,
    variables: dict[str, typing.Any] = {},
    num_timeout_retries: int = 0,
) -> typing.Any:
    if num_timeout_retries < 0:
        raise ValueError("num_timeout_retries must be >= 0")
    for attempt_no in range(num_timeout_retries + 1):
        try:
            return wandb_public_api().client.execute(  # type: ignore
                gql(query_str),
                variable_values=variables,
            )
        except exceptions.Timeout as e:
            if attempt_no == num_timeout_retries:
                raise
            logging.warn(
                f'wandb GQL query timed out: "{e}", retrying (num_attempts={attempt_no + 1})'
            )


def introspect_server_schema(num_timeout_retries: int = 0) -> GraphQLSchema:
    introspection_query = graphql.get_introspection_query()
    payload = query_with_retry(introspection_query, {}, num_timeout_retries)
    return graphql.build_client_schema(payload)


def wandb_gql_query(
    query_str: str,
    variables: dict[str, typing.Any] = {},
    num_timeout_retries: int = 0,
) -> typing.Any:
    return query_with_retry(query_str, variables, num_timeout_retries)


def set_wandb_thread_local_api_settings(
    api_key: typing.Optional[str],
    cookies: typing.Optional[typing.Dict],
    headers: typing.Optional[typing.Dict],
) -> None:
    _thread_local_api_settings.api_key = api_key
    _thread_local_api_settings.cookies = cookies
    _thread_local_api_settings.headers = headers


class WandbThreadLocalApiSettings(typing.TypedDict):
    api_key: typing.Optional[str]
    cookies: typing.Optional[dict]
    headers: typing.Optional[dict]


def reset_wandb_thread_local_api_settings() -> None:
    _thread_local_api_settings.api_key = None
    _thread_local_api_settings.cookies = None
    _thread_local_api_settings.headers = None


_WANDB_CLIENT_PATCHED = False


def do_patch() -> None:
    global _WANDB_CLIENT_PATCHED
    if _WANDB_CLIENT_PATCHED:
        return

    ### Start _parse_artifact_path Patch ###
    # Can be removed after https://github.com/wandb/wandb/pull/6083 is merged and deployed.
    orig_parse_artifact_path = public.Api._parse_artifact_path

    def new_parse_artifact_path(self, path):  # type: ignore
        # Adding this short circuit skips the potential access of
        # self.default_entity which incurs a network call. However, if the path
        # is fully qualified, then this `entity` is thrown away.
        parts = [] if path is None else path.split("/")
        if len(parts) == 3:
            return parts
        return orig_parse_artifact_path(self, path)

    public.Api._parse_artifact_path = new_parse_artifact_path  # type: ignore
    ### End _parse_artifact_path Patch ###

    _WANDB_CLIENT_PATCHED = True


do_patch()



import pytest
from PIL import Image

from weave.legacy.weave import api as weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave import ops_arrow

from ... import errors
from weave.legacy.tests.util import geom


def test_mapped_method_on_custom_type():
    segment = geom.LineSegment(0.0, 0.1, 0.9, 1.1)

    len = weave.use(segment.length())
    assert len == 1.81

    mid = weave.use(segment.midpoint())
    assert mid.x == 0.45
    assert mid.y == 0.6

    segments = ops_arrow.to_weave_arrow(
        [
            geom.LineSegment(0.0, 0.1, 0.9, 1.1),
            geom.LineSegment(0.0, 0.2, 0.4, 1.1),
        ]
    )
    lens = weave.use(segments.map(lambda seg: seg.length()))
    assert lens[0] == 1.81
    assert lens[1] == pytest.approx(0.97)

    # mid = weave.use(segments.midpoint())
    # mid = weave.use(segments.map(lambda seg: seg.midpoint()))
    # assert mid[0].x == 0.45
    # assert mid[0].y == 0.6
    # assert mid[1].x == 0.2
    # assert mid[1].y == pytest.approx(0.65)


def test_mapped_method_returning_custom_type():
    segments = ops_arrow.to_weave_arrow(
        [
            geom.LineSegment(0.0, 0.1, 0.9, 1.1),
            geom.LineSegment(0.0, 0.2, 0.4, 1.1),
        ]
    )

    mid = weave.use(segments.map(lambda seg: seg.midpoint()))

    assert mid[0].x == 0.45
    assert mid[0].y == 0.6
    assert mid[1].x == 0.2
    assert mid[1].y == pytest.approx(0.65)


def test_mapped_pick():
    data = [
        {"a": 5, "b": 9},
        {"a": 6, "b": 10},
    ]
    arrow_arr = ops_arrow.to_arrow(data)

    assert weave.use(ops_arrow.dict.pick(arrow_arr, "b")).to_pylist_raw() == [9, 10]


@pytest.mark.skip("disabled decorator_type() constructor")
def test_constructor():
    expected = geom.Point2d(0.5, 0.6)
    point2d_node = geom.Point2d.constructor(x=0.5, y=0.6)
    assert weave.use(point2d_node) == expected

    with pytest.raises(errors.WeaveDispatchError):
        weave.use(geom.Point2d.constructor(x=0.5))



"""
This file contains the interface definition for the Trace Server Query model. It
is heavily inspired by the MongoDB query language, but is a subset of the full
MongoDB query language. In particular, we have made the following
simplifications:

* We only support the "aggregation" operators, not the "query" operators. This is
    purely for simplicity and because the "aggregation" operators are more powerful.
    The Mongo docs language has evolved over time and the primary query language
    is column-oriented. However, the more expressive aggregation language can be
    used for both direct queries, but also for column comparison and
    calculations. We can add support for the "query" operators in the future if
    needed.

* We only support a subset of the operators / shorthand forms for now. We can add
    more operators in the future as needed.

    * One notable omission here is the lack of support for "$field" as a shorthand for
        the "getField"  operator.

* We have _added_ a `$contains` operator which is not in the MongoDB query
    language. This is a simple substring match operator.
"""

import typing

from pydantic import BaseModel, Field

# Operations: all operations have the form of a single property
# with the name of the operation suffixed with an underscore.
# Subset of Mongo _Aggregation_ Operators: https://www.mongodb.com/docs/manual/reference/operator/aggregation/
# Starting with these operators for now since they are the most common and with negation
# can cover most of the other operators.


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/literal/
# Can be any standard json-able value
class LiteralOperation(BaseModel):
    literal_: typing.Union[
        str,
        int,
        float,
        bool,
        dict[str, "LiteralOperation"],
        list["LiteralOperation"],
        None,
    ] = Field(alias="$literal")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/getField/
# Note: currently only the shorthand form is supported.
# Field should be a key of `CallSchema`. For dictionary fields (`attributes`,
# `inputs`, `outputs`, `summary`), the field can be dot-separated.
# Here we support "dot notation" for accessing nested fields:
# https://www.mongodb.com/docs/manual/core/document/#dot-notation
class GetFieldOperator(BaseModel):
    # Tim: We will likely want to revisit this before making it public. Here are some concerns:
    # 1. Mongo explicitly says that `getField` is used to access fields without dot notation - this
    #    is not how we are handling it here - we are using dot notation - this could be resolved by
    #    supporting the `$field.with.path` shorthand.
    # 2. As Jamie pointed out, the parsing of this field is not very robust and susceptible to issues when:
    #    - The field part name contains a dot
    #    - The field part name is a valid integer (currently interpreted as a list index)
    #    - The field part name contains a double quote (will result in failed lookup - see `_quote_json_path` in `clickhouse_trace_server_batched.py`)
    #    These issues could be resolved by using an alternative syntax (perhaps backticks, square brackets, etc.). However
    #    this would diverge from the current Mongo syntax.
    get_field_: str = Field(alias="$getField")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/convert/
class ConvertOperation(BaseModel):
    convert_: "ConvertSpec" = Field(alias="$convert")


CastTo = typing.Literal["double", "string", "int", "bool", "exists"]


class ConvertSpec(BaseModel):
    input: "Operand"
    # Subset of https://www.mongodb.com/docs/manual/reference/bson-types/#std-label-bson-types
    to: CastTo


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/and/
class AndOperation(BaseModel):
    and_: typing.List["Operand"] = Field(alias="$and")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/or/
class OrOperation(BaseModel):
    or_: typing.List["Operand"] = Field(alias="$or")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/not/
class NotOperation(BaseModel):
    not_: typing.Tuple["Operand"] = Field(alias="$not")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/eq/
class EqOperation(BaseModel):
    eq_: typing.Tuple["Operand", "Operand"] = Field(alias="$eq")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/gt/
class GtOperation(BaseModel):
    gt_: typing.Tuple["Operand", "Operand"] = Field(alias="$gt")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/gte/
class GteOperation(BaseModel):
    gte_: typing.Tuple["Operand", "Operand"] = Field(alias="$gte")


# https://www.mongodb.com/docs/manual/reference/operator/aggregation/in/
class InOperation(BaseModel):
    in_: typing.Tuple["Operand", list["Operand"]] = Field(alias="$in")


# This is not technically in the Mongo spec. Mongo has:
# https://www.mongodb.com/docs/manual/reference/operator/aggregation/regexMatch/,
# however, rather than support a full regex match right now, we will
# support a substring match. We can add regex support later if needed.
class ContainsOperation(BaseModel):
    contains_: "ContainsSpec" = Field(alias="$contains")


class ContainsSpec(BaseModel):
    input: "Operand"
    substr: "Operand"
    case_insensitive: typing.Optional[bool] = False


# Convenience type for all Operands and Operations
Operation = typing.Union[
    AndOperation,
    OrOperation,
    NotOperation,
    EqOperation,
    GtOperation,
    GteOperation,
    InOperation,
    ContainsOperation,
]
Operand = typing.Union[
    LiteralOperation, GetFieldOperator, ConvertOperation, "Operation"
]


# Update the models to include the recursive types
LiteralOperation.model_rebuild()
GetFieldOperator.model_rebuild()
AndOperation.model_rebuild()
OrOperation.model_rebuild()
NotOperation.model_rebuild()
EqOperation.model_rebuild()
GtOperation.model_rebuild()
GteOperation.model_rebuild()
InOperation.model_rebuild()
ContainsOperation.model_rebuild()


class Query(BaseModel):
    # Here, we use `expr_` to match the MongoDB query language's "aggregation" operator syntax.
    # This is certainly a subset of the full MongoDB query language, but it is a good starting point.
    # https://www.mongodb.com/docs/manual/reference/operator/query/expr/#mongodb-query-op.-expr
    expr_: Operation = Field(alias="$expr")
    # In the future, we could have other top-level Query Operators as described here:
    # https://www.mongodb.com/docs/manual/reference/operator/query/



from weave.legacy.weave import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from weave.legacy.weave.panels_py import (
    generator_templates,
    panel_autoboard,
    panel_llm_monitor,
    panel_observability,
    panel_seedboard,
    panel_trace_monitor,
)

# This is just an example, uncomment to enable
# from weave.legacy.weave.panels_py import panel_example_custom_board

_context_state.clear_loading_built_ins(_loading_builtins_token)



import typing

if typing.TYPE_CHECKING:
    from weave.legacy.weave import weave_types as types
    from weave.legacy.weave import artifact_local


class Mapper:
    def __init__(self, type_, mapper, artifact, path):
        self.type = type_

    def result_type(self):
        raise NotImplementedError

    def apply(self, obj) -> typing.Union[dict, typing.Any]:
        return obj


def make_mapper(map_fn):
    def mapper(type_, artifact, path=[], mapper_options=None):
        return map_fn(type_, mapper, artifact, path, mapper_options=mapper_options)

    return mapper



import datetime
import hashlib
import json
import typing

import wandb
from langchain.callbacks.tracers import wandb as LCW

# We should move this logic into our built-in langchain integration
from langchain.callbacks.tracers.base import BaseTracer
from langchain.callbacks.tracers.schemas import Run
from wandb.sdk.data_types import trace_tree

from weave.legacy.weave import stream_data_interfaces
from weave.legacy.weave.monitoring import StreamTable


def _hash_id(s: str) -> str:
    return hashlib.md5(s.encode("utf-8")).hexdigest()[:16]


patched = False


# This should be committed to the langchain repo
def patch_serialize() -> None:
    global patched
    if patched:
        return
    patched = True

    old_serialize_io = LCW._serialize_io

    def new_serialize_io(run_inputs):  # type: ignore
        if run_inputs is None:
            return {}
        return old_serialize_io(run_inputs)

    LCW._serialize_io = new_serialize_io


patch_serialize()


class WeaveTracer(BaseTracer):
    def __init__(self, stream_uri: str, **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)
        self.run_processor = LCW.RunProcessor(wandb, trace_tree)
        self._st = StreamTable(stream_uri)

    def _persist_run(self, run: Run) -> None:
        root_span = self.run_processor.process_span(run)
        if root_span is None:
            return
        model_dict = self.run_processor.process_model(run)
        model_str = json.dumps(model_dict)
        root_span.attributes["model"] = {
            # This is hardly an "id", since it is unique for each run.
            # Need to reconcile this with LC
            "id": _hash_id(model_str),
            "obj": model_str,
        }
        spans = stream_data_interfaces.wb_span_to_weave_spans(root_span)
        for span in spans:
            self._st.log(span)
